# -*- coding: utf-8 -*-

import re
from collections import OrderedDict
from .column import Column
from .foreign_key_constraint import ForeignKeyConstraint
from .schema_manager import SchemaManager
from .platforms.mysql_platform import MySQLPlatform


class MySQLSchemaManager(SchemaManager):
    def _get_portable_table_column_definition(self, table_column):
        db_type = table_column["type"].lower()
        type_match = re.match("(.+)\((.*)\).*", db_type)
        if type_match:
            db_type = type_match.group(1)

        if "length" in table_column:
            length = table_column["length"]
        else:
            if type_match and type_match.group(2) and "," not in type_match.group(2):
                length = int(type_match.group(2))
            else:
                length = 0

        fixed = None

        if "name" not in table_column:
            table_column["name"] = ""

        precision = None
        scale = None
        extra = {}

        type = self._platform.get_type_mapping(db_type)

        if db_type in ["char", "binary"]:
            fixed = True
        elif db_type in ["float", "double", "real", "decimal", "numeric"]:
            match = re.match("([A-Za-z]+\(([0-9]+),([0-9]+)\))", table_column["type"])
            if match:
                precision = match.group(1)
                scale = match.group(2)
                length = None
        elif db_type == "tinytext":
            length = MySQLPlatform.LENGTH_LIMIT_TINYTEXT
        elif db_type == "text":
            length = MySQLPlatform.LENGTH_LIMIT_TEXT
        elif db_type == "mediumtext":
            length = MySQLPlatform.LENGTH_LIMIT_MEDIUMTEXT
        elif db_type == "tinyblob":
            length = MySQLPlatform.LENGTH_LIMIT_TINYBLOB
        elif db_type == "blob":
            length = MySQLPlatform.LENGTH_LIMIT_BLOB
        elif db_type == "mediumblob":
            length = MySQLPlatform.LENGTH_LIMIT_MEDIUMBLOB
        elif db_type in ["tinyint", "smallint", "mediumint", "int", "bigint", "year"]:
            length = None
        elif db_type == "enum":
            length = None
            extra["definition"] = "({})".format(type_match.group(2))

        if length is None or length == 0:
            length = None

        options = {
            "length": length,
            "unsigned": table_column["type"].find("unsigned") != -1,
            "fixed": fixed,
            "notnull": table_column["null"] != "YES",
            "default": table_column.get("default"),
            "precision": None,
            "scale": None,
            "autoincrement": table_column["extra"].find("auto_increment") != -1,
            "extra": extra,
        }

        if scale is not None and precision is not None:
            options["scale"] = scale
            options["precision"] = precision

        column = Column(table_column["field"], type, options)

        if "collation" in table_column:
            column.set_platform_option("collation", table_column["collation"])

        return column

    def _get_portable_table_indexes_list(self, table_indexes, table_name):
        new = []
        for v in table_indexes:
            v = dict((k.lower(), value) for k, value in v.items())
            if v["key_name"] == "PRIMARY":
                v["primary"] = True
            else:
                v["primary"] = False

            if "FULLTEXT" in v["index_type"]:
                v["flags"] = {"FULLTEXT": True}
            else:
                v["flags"] = {"SPATIAL": True}

            new.append(v)

        return super(MySQLSchemaManager, self)._get_portable_table_indexes_list(
            new, table_name
        )

    def _get_portable_table_foreign_keys_list(self, table_foreign_keys):
        foreign_keys = OrderedDict()

        for value in table_foreign_keys:
            value = dict((k.lower(), v) for k, v in value.items())
            name = value.get("constraint_name", "")

            if name not in foreign_keys:
                if "delete_rule" not in value or value["delete_rule"] == "RESTRICT":
                    value["delete_rule"] = ""

                if "update_rule" not in value or value["update_rule"] == "RESTRICT":
                    value["update_rule"] = ""

                foreign_keys[name] = {
                    "name": name,
                    "local": [],
                    "foreign": [],
                    "foreign_table": value["referenced_table_name"],
                    "on_delete": value["delete_rule"],
                    "on_update": value["update_rule"],
                }

            foreign_keys[name]["local"].append(value["column_name"])
            foreign_keys[name]["foreign"].append(value["referenced_column_name"])

        result = []
        for constraint in foreign_keys.values():
            result.append(
                ForeignKeyConstraint(
                    constraint["local"],
                    constraint["foreign_table"],
                    constraint["foreign"],
                    constraint["name"],
                    {
                        "on_delete": constraint["on_delete"],
                        "on_update": constraint["on_update"],
                    },
                )
            )

        return result



import threading 

## print("load Thread")

class ThreadObj():
    def __init__(self, threadobj:threading.Thread) -> None:
        self.threadobj = threadobj 
    
    def Join(self):
        self.threadobj.join()

class ThreadObjs():
    def __init__(self, threadobjs:list[threading.Thread]) -> None:
        self.threadobjs = threadobjs
    
    def Join(self):
        [i.join() for i in self.threadobjs]

def Thread(func, *args, count:int=1) -> ThreadObjs | ThreadObj:
    if count == 1:
        t = threading.Thread(target=func, args=args)
        t.daemon = True 
        t.start()

        return ThreadObj(t)
    elif count > 1:
        ts = []
        for _ in range(count):
            t = threading.Thread(target=func, args=args)
            t.daemon = True 
            t.start()

            ts.append(t)

        return ThreadObjs(ts)
    else:
        raise Exception("count异常")

if __name__ == "__main__":
    import time 

    def p(s:str, ss:str):
        while True:
            time.sleep(1)
            print(s, ss, time.time())

    t = Thread(p, "oo", "kk")

    while True:
        time.sleep(1)






from packaging import version

def VersionCompare(version1:str, operator:str, version2:str) -> bool:
    v1 = version.parse(version1)
    v2 = version.parse(version2)
    o = operator

    if o == ">":
        return v1 > v2 
    elif o == ">=":
        return v1 >= v2 
    elif o == "=" or o == "==":
        return v1 == v2
    elif o == "<":
        return v1 < v2 
    elif o == "<=":
        return v1 <= v2 
    else:
        raise Exception(f'Unsupport operator: {o}, must be on of >, >=, =, ==, <, <=')


import prometheus_client as pc

class PrometheusCounter():
    def __init__(self, name:str, help:str, registry:pc.CollectorRegistry=None) -> None:
        if registry != None:
            self.c = pc.Counter(name, help, registry=registry)
        else:
            self.c = pc.Counter(name, help)
        self.current = 0

        # import ipdb
        # ipdb.set_trace()
    
    def Add(self, num:int|float=1):
        self.c.inc(num)
    
    def Set(self, num:int|float=1):
        if num < self.current:
            raise Exception(f"Count类型只能设置更大类型的值, 当前值和想设置的值为: {self.current}, {num}")
        
        if num == self.current:
            return 

        self.c.inc(num - self.current)
        self.current = num

class PrometheusCounterVec():
    def __init__(self, name:str, labels:list[str], help:str, registry:pc.CollectorRegistry=None) -> None:
        self.labels = labels 
        if registry != None:
            self.c = pc.Counter(name, help, labels, registry=registry)
        else:
            self.c = pc.Counter(name, help, labels)
        self.current = {}
    
    def Add(self, labels:dict|list, num:int|float=1):
        """
        It adds a new label to the metric.
        
        :param labels: a list of labels, or a dict of labels
        :type labels: dict|list
        :param num: The number to increment the counter by, defaults to 1
        :type num: int|float (optional)
        """
        if type(labels) == dict:
            lb = []
            for k in self.labels:
                if k in labels:
                    lb.append(labels[k])
                else:
                    lb.append("")
        elif type(labels) == list:
            if len(self.labels) == len(labels):
                lb = labels
            else:
                lb = labels[:len(self.labels)] + [""]*(len(self.labels) - len(labels))

        lbr = repr(lb)
        if lbr not in self.current:
            self.current[lbr] = 0

        self.current[lbr] = self.current[lbr] + num
        self.c.labels(*lb).inc(num)
    
    def Set(self, labels:dict|list, num:int|float=1):
        """
        It adds a new label to the metric.
        
        :param labels: a list of labels, or a dict of labels
        :type labels: dict|list
        :param num: 需要设置到counter的数字, 只能等于大于之前的数字, 否则抛异常
        :type num: int|float (optional)
        """
        if type(labels) == dict:
            lb = []
            for k in self.labels:
                if k in labels:
                    lb.append(labels[k])
                else:
                    lb.append("")
        elif type(labels) == list:
            if len(self.labels) == len(labels):
                lb = labels
            else:
                lb = labels[:len(self.labels)] + [""]*(len(self.labels) - len(labels))

        lbr = repr(lb)
        if lbr not in self.current:
            self.current[lbr] = 0
        
        if num < self.current[lbr]:
            raise Exception(f"Count类型只能设置更大类型的值, 当前值和想设置的值为: {self.current}, {num}")
        
        if num == self.current[lbr]:
            return 

        self.c.labels(*lb).inc(num - self.current[lbr])
        self.current[lbr] = num

class PrometheusGauge:
    def __init__(self, name:str, help:str, registry:pc.CollectorRegistry=None) -> None:
        if registry != None:
            self.g = pc.Gauge(name, help, registry=registry)
        else:
            self.g = pc.Gauge(name, help)
    
    def Set(self, num:int|float):
        self.g.set(num)

class PrometheusGaugeVec():
    def __init__(self, name:str, labels:list[str], help:str, registry:pc.CollectorRegistry=None) -> None:
        self.labels = labels 
        if registry != None:
            self.g = pc.Gauge(name, help, labels, registry=registry)
        else:
            self.g = pc.Gauge(name, help, labels)
    
    def Set(self, labels:dict|list, num:int|float):
        """
        It adds a number to the graph.
        
        :param labels: The labels of the histogram
        :type labels: dict|list
        :param num: The number of times the label is added, defaults to 1
        :type num: int|float (optional)
        """
        if type(labels) == dict:
            lb = []
            for k in self.labels:
                if k in labels:
                    lb.append(labels[k])
                else:
                    lb.append("")
        elif type(labels) == list:
            if len(self.labels) == len(labels):
                lb = labels
            else:
                lb = labels[:len(self.labels)] + [0]*(len(self.labels) - len(labels))
        self.g.labels(*lb).set(num)



from __future__ import annotations

#print("load " + '/'.join(__file__.split('/')[-2:]))

import argparse

class Argparser():
    def __init__(self, description:str=None) -> None:
        self.parser = argparse.ArgumentParser(description=description)

    def Add(self, arg:str, help:str=None) -> Argparser:
        return self.AddString(arg, help)
    
    def AddOpt(self, arg:str, default:str=None, help:str=None) -> Argparser:
        return self.AddOptString(arg, default, help)
    
    def AddStr(self, arg:str, help:str=None) -> Argparser:
        return self.AddString(arg, help)
    
    def AddOptStr(self, arg:str, default:str=None, help:str=None) -> Argparser:
        return self.AddOptString("--" + arg, default, help)
    
    def AddString(self, arg:str, help:str=None) -> Argparser:
        self.parser.add_argument(arg, help=help, type=str)
        return self
    
    def AddOptString(self, arg:str, default:str=None, help:str=None) -> Argparser:
        self.parser.add_argument("--" + arg, default=default, help=help, type=str)
        return self

    def AddOptBool(self, arg:str, default:bool=False, help:str=None) -> Argparser:
        self.parser.add_argument("--" + arg, default=default, help=help, action='store_true')
        return self
    
    def AddInt(self, arg:str, help:str=None) -> Argparser:
        self.parser.add_argument(arg, help=help, type=int)
        return self
    
    def AddOptInt(self, arg:str, default:int=None, help:str=None) -> Argparser:
        self.parser.add_argument("--" + arg, default=default, help=help, type=int)
        return self
    
    def AddFloat(self, arg:str, help:str=None) -> Argparser:
        self.parser.add_argument(arg, help=help, type=float)
        return self
    
    def AddOptFloat(self, arg:str, default:float=None, help:str=None) -> Argparser:
        self.parser.add_argument("--" + arg, default=default, help=help, type=float)
        return self
    
    def Get(self):
        return self.parser.parse_args()

if __name__ == "__main__":
    args = (
        Argparser().
        AddOptBool("arg1").
        AddString("arg2"). 
        Add("arg3", "help string 3"). 
        AddOpt("optionArgs4"). 
        AddOpt("optionArgs5", "defaultValue5").
        AddOpt("optionArgs6", "defaultValue6", "help string 6"). 
        AddInt("intKey"). 
        AddOptFloat("floatKey").
        Get()
    )
    print(args.arg3)
    print(args.floatKey)
    print(args)



from typing import TYPE_CHECKING
from lazy_imports import LazyImporter
import sys

_import_structure = {
    "src": [
        "Dumps",
        "Loads", 
        "ExtraValueByKey",
        "Valid"
    ],
}

if TYPE_CHECKING:
    from .src import (
        Dumps,
        Loads,
        ExtraValueByKey,
        Valid
    )
else:
    sys.modules[__name__] = LazyImporter(
        __name__,
        globals()["__file__"],
        _import_structure,
        extra_objects={},
    )





# -*- coding: utf-8 -*-

from .install_command import InstallCommand
from .migrate_command import MigrateCommand
from .make_command import MigrateMakeCommand
from .rollback_command import RollbackCommand
from .status_command import StatusCommand
from .reset_command import ResetCommand
from .refresh_command import RefreshCommand



# -*- coding: utf-8 -*-

from contextlib import contextmanager
from ...query.expression import QueryExpression
from ..collection import Collection
from ..builder import Builder


class Relation(object):

    _constraints = True

    def __init__(self, query, parent):
        """
        :param query: A Builder instance
        :type query: orm.orator.Builder

        :param parent: The parent model
        :type parent: Model
        """
        self._query = query
        self._parent = parent
        self._related = query.get_model()
        self._extra_query = None

        self.add_constraints()

    def add_constraints(self):
        """
        Set the base constraints on the relation query.

        :rtype: None
        """
        raise NotImplementedError

    def add_eager_constraints(self, models):
        """
        Set the constraints for an eager load of the relation.

        :type models: list
        """
        raise NotImplementedError

    def init_relation(self, models, relation):
        """
        Initialize the relation on a set of models.

        :type models: list
        :type relation:  str
        """
        raise NotImplementedError

    def match(self, models, results, relation):
        """
        Match the eagerly loaded results to their parents.

        :type models: list
        :type results: Collection
        :type relation:  str
        """
        raise NotImplementedError

    def get_results(self):
        """
        Get the results of the relationship.
        """
        raise NotImplementedError

    def get_eager(self):
        """
        Get the relationship for eager loading.

        :rtype: Collection
        """
        return self.get()

    def touch(self):
        """
        Touch all of the related models for the relationship.
        """
        column = self.get_related().get_updated_at_column()

        self.raw_update({column: self.get_related().fresh_timestamp()})

    def raw_update(self, attributes=None):
        """
        Run a raw update against the base query.

        :type attributes: dict

        :rtype: int
        """
        if attributes is None:
            attributes = {}

        if self._query is not None:
            return self._query.update(attributes)

    def get_relation_count_query(self, query, parent):
        """
        Add the constraints for a relationship count query.

        :type query: Builder
        :type parent: Builder

        :rtype: Builder
        """
        query.select(QueryExpression("COUNT(*)"))

        key = self.wrap(self.get_qualified_parent_key_name())

        return query.where(self.get_has_compare_key(), "=", QueryExpression(key))

    @classmethod
    @contextmanager
    def no_constraints(cls, with_subclasses=False):
        """
        Runs a callback with constraints disabled on the relation.
        """
        cls._constraints = False

        if with_subclasses:
            for klass in cls.__subclasses__():
                klass._constraints = False

        try:
            yield cls
        except Exception:
            raise
        finally:
            cls._constraints = True
            if with_subclasses:
                for klass in cls.__subclasses__():
                    klass._constraints = True

    def get_keys(self, models, key=None):
        """
        Get all the primary keys for an array of models.

        :type models: list
        :type key: str

        :rtype: list
        """
        return list(
            set(
                map(
                    lambda value: value.get_attribute(key) if key else value.get_key(),
                    models,
                )
            )
        )

    def get_query(self):
        return self._query

    def get_base_query(self):
        return self._query.get_query()

    def merge_query(self, query):
        if isinstance(query, Builder):
            query = query.get_query()

        self._query.merge(query)

    def get_parent(self):
        return self._parent

    def get_qualified_parent_key_name(self):
        return self._parent.get_qualified_key_name()

    def get_related(self):
        return self._related

    def created_at(self):
        """
        Get the name of the "created at" column.

        :rtype: str
        """
        return self._parent.get_created_at_column()

    def updated_at(self):
        """
        Get the name of the "updated at" column.

        :rtype: str
        """
        return self._parent.get_updated_at_column()

    def get_related_updated_at(self):
        """
        Get the name of the related model's "updated at" column.

        :rtype: str
        """
        return self._related.get_updated_at_column()

    def wrap(self, value):
        """
        Wrap the given value with the parent's query grammar.

        :rtype: str
        """
        return self._parent.new_query().get_query().get_grammar().wrap(value)

    def set_parent(self, parent):
        self._parent = parent

    def set_extra_query(self, query):
        self._extra_query = query

    def new_query(self, related=None):
        if related is None:
            related = self._related

        query = related.new_query()

        if self._extra_query:
            query.merge(self._extra_query.get_query())

        return query

    def new_instance(self, model, **kwargs):
        new = self._new_instance(model, **kwargs)

        if self._extra_query:
            new.set_extra_query(self._extra_query)

        return new

    def __dynamic(self, method):
        attribute = getattr(self._query, method)

        def call(*args, **kwargs):
            result = attribute(*args, **kwargs)

            if result is self._query:
                return self

            return result

        if not callable(attribute):
            return attribute

        return call

    def __getattr__(self, item):
        return self.__dynamic(item)



import tweepy
import typing 

tweetFields = ['author_id', 'created_at', 'geo', 'id', 'lang', 'text']

#print("load " + '/'.join(__file__.split('/')[-2:]))

class EssentialTweet():
    def __init__(self) -> None:
        self.ID:int = None
        self.Time:int = None 
        self.Text:str = None 
        self.Language:str = None 
    
    def __repr__(self) -> str:
        return f"EssentialTweet(ID={self.ID} Time={self.Time} Language={self.Language} Text={self.Text})"
    
    def __str__(self) -> str:
        return f"EssentialTweet(ID={self.ID} Time={self.Time} Language={self.Language} Text={self.Text})"

class Essential():
    def __init__(self, bearerToken:str) -> None:
        self.api = tweepy.Client(bearer_token=bearerToken, wait_on_rate_limit=True)

    def _wrapStatus(self, status) -> EssentialTweet:
        t = EssentialTweet()

        t.ID = status.id 
        t.Time = int(status.created_at.timestamp())

        t.Text = status.text

        t.Language = status.lang

        return t
    
    def Search(self, keyword:str, sinceID:int=None, tweetPerRequest:int=10) -> typing.Iterable[EssentialTweet]:
        tweets = tweepy.Paginator(self.api.search_recent_tweets, query=keyword, since_id=sinceID, tweet_fields=tweetFields, max_results=tweetPerRequest).flatten()
        for status in tweets:
            yield self._wrapStatus(status)
    
    def Timeline(self, screename:str, sinceID:int=None, tweetPerRequest:int=10) -> typing.Iterable[EssentialTweet]:
        u = self.api.get_user(username=screename)
        if len(u.errors) != 0:
            raise Exception("User not exists: " + screename)

        tweets = tweepy.Paginator(self.api.get_users_tweets, id=u.data.id, since_id=sinceID, tweet_fields=tweetFields, max_results=tweetPerRequest).flatten()
        for status in tweets:
            yield self._wrapStatus(status)

if __name__ == "__main__":
    t = Essential(twitterBearerToken)

    t.Timeline("EtherChecker")

    count = 0
    for tt in t.Timeline("EtherChecker"):
        print(tt)
        count += 1
        if count > 30:
            break 


"""LanguageResponse"""


class LanguageResponse:
    """LanguageResponse"""

    def __init__(self, language: str, name: str = None):
        """

        :param language: 语言代码
        :param name: 语言名字
        """

        self.language = language
        self.name = name

    def __repr__(self):
        return self.__class__.__qualname__ + f'(language={repr(self.language)}, name={repr(self.language)})'



from ..... import Http

import json
import uuid
import time

#print("load " + '/'.join(__file__.split('/')[-2:]))

url = 'https://www.binance.com/bapi/composite/v1/public/official-channel/verify'

def Twitter(account:str, waiteOnRateLimit:bool=True) -> bool:
    tu = f"https://twitter.com/{account}"
    data = {"content": tu}

    while True:
        headers = {
            "bnc-uuid": str(uuid.uuid4()),
            "content-type": "application/json",
        }
        
        resp = Http.PostRaw(url, json.dumps(data), headers=headers)

        if resp.StatusCode == 200:
            break 

        elif resp.StatusCode == 429:
            if waiteOnRateLimit == True:
                time.sleep(30)
            else:
                raise Exception(resp)

    c = json.loads(resp.Content)

    if len(c["data"]['data']) == 0:
        return False 
    
    for d in c['data']['data']:
        if d['content'] == tu:
            return True 
    
    return False 



from __future__ import annotations

# 需要配合以下服务食用
# 
# version: '3'
# services:
#   matrix_bot:
#     image: darren2046/matrix-bot:0.0.12
#     container_name: matrix-bot
#     restart: always
#     #ports:
#     #   - "8888:8888" 
#     environment:
#       MATRIX_SERVER: "https://your.homeserver.com"
#       MATRIX_USER: account_username 
#       MATRIX_PASS: account_password
#       API_PASS: password_for_call_this_api_server # can be empty
#     dns:
#       - 8.8.8.8
#       - 4.4.4.4
#     volumes:
#       - /data/cr-volumes/matrix-bot/data:/data
# 备注:
# 版本0.0.1
#     arm64的可以调用http api发送消息, 也可以收消息(发送id给bot返回房间号)
#     amd64的可以调用http api发送消息, 但是不能收消息, 一脸蒙逼
#     

#print("load " + '/'.join(__file__.split('/')[-2:]))

from .. import Http
from .. import Base64
from .. import Json

import requests
import time

def retryOnNetworkError(func): # func是被包装的函数
    def ware(self, *args, **kwargs): # self是类的实例
        while True:
            try:
                res = func(self, *args, **kwargs)
                break
            except requests.exceptions.ReadTimeout as e:
                time.sleep(1)

        return res
    
    return ware

class MatrixBotMessage():
    def __init__(self, mb:MatrixBot, time:int, text:str, user:str, room:str) -> None:
        self.mb = mb 
        self.Time = time 
        self.Text = text 
        self.User = user 
        self.Room = room
    
    @retryOnNetworkError
    def Reply(self, message:str):
        resp = Http.PostForm(
            self.mb.apiserver + "/matrix-bot/send/text", 
            {
                "room": self.Room, 
                'text': message,
                'password': self.mb.password,
            })
        if resp.StatusCode != 200:
            raise Exception("发送消息错误:", resp.StatusCode)
    
    @retryOnNetworkError
    def ReplyImage(self, path:str):
        resp = Http.PostForm(
            self.mb.apiserver + "/matrix-bot/send/image", 
            {
                "room": self.Room, 
                'image': Base64.Encode(open(path, 'rb').read()),
                'password': self.mb.password,
            })
        if resp.StatusCode != 200:
            raise Exception("发送消息错误:", resp.StatusCode)
    
    def __repr__(self):
        return f"MatrixBotMessage(Time={self.Time} Room={self.Room} User={self.User} Text={self.Text})"
    
    def __str__(self):
        return f"MatrixBotMessage(Time={self.Time} Room={self.Room} User={self.User} Text={self.Text})"

class MatrixBot():
    def __init__(self, apiserver:str, password:str="") -> None:
        self.apiserver = apiserver.rstrip('/')
        self.password = password 
        self.room = None 

        if not self.apiserver.startswith("http://") and not self.apiserver.startswith("https://"):
            self.apiserver = "https://" + self.apiserver
    
    @retryOnNetworkError
    def SetRoom(self, room:str) -> MatrixBot:
        """
        如果room的id是 !abcdefghiljkmn:example.com, 那么room可以是abcdefghiljkmn, 默认取homeserver的域名
        
        :param room: The room you want to join
        :type room: str
        """
        self.room = room

        return self
    
    @retryOnNetworkError
    def Send(self, message:str, msgtype:str="text"):
        """
        > Send a message to the room. 
        
        The function takes two parameters: 
        
        - `message`: The message to be sent. 
        - `msgtype`: The type of the message. 
        
        The `msgtype` parameter can be one of the following: 
        
        - `text`: Plain text. 
        - `markdown`: Markdown text. 
        - `md`: Markdown text. 
        
        The function will raise an exception if the room is not set. 
        
        The function will raise an exception if the message cannot be sent.
        
        :param message: The message to be sent
        :type message: str
        :param msgtype: The type of message to be sent. Currently only supports text and markdown,
        defaults to text
        :type msgtype: str (optional)
        """
        if self.room == None:
            raise Exception("Need to set room first. ")
        
        if msgtype == "text":
            url = "/matrix-bot/send/text"
        else:
            url = "/matrix-bot/send/markdown"

        resp = Http.PostForm(
            self.apiserver + url, 
            {
                "room": self.room, 
                'text': message,
                'password': self.password,
            })
        if resp.StatusCode != 200:
            raise Exception("发送消息错误:", resp.StatusCode)
    
    @retryOnNetworkError
    def SendImage(self, path:str):
        if self.room == None:
            raise Exception("Need to set room first. ")

        resp = Http.PostForm(
            self.apiserver + "/matrix-bot/send/image", 
            {
                "room": self.room, 
                'image': Base64.Encode(open(path, 'rb').read()),
                'password': self.password,
            })
        if resp.StatusCode != 200:
            raise Exception("发送消息错误:", resp.StatusCode)
    
    @retryOnNetworkError
    def GetMessage(self, num:int=10, room:str=None) -> list[MatrixBotMessage]:
        """
        > Get the last 10 messages from the current room, will get messages from all rooms if the current room is not set. 
        > If the room is like !abcdefg:example.com and homeserver is example.com, then only need to set 'abcdefg' as room id. 
        > If the room is set to 'all', will get messages from all rooms.
        
        :param num: The number of messages to get, defaults to 10
        :type num: int (optional)
        :param room: The room to send the message to. If not specified, the default room is used
        :type room: str
        :return: A list of MatrixBotMessage objects.
        """
        if room == None:
            if self.room == None:
                room = 'all'
            else:
                room = self.room 

        res = Http.Get(self.apiserver + "/matrix-bot/get/message", {'password': self.password, 'num': str(num), "room": room}).Content
        # Lg.Trace(res)
        res = Json.Loads(res)
        resm = []
        for r in res:
            resm.append(MatrixBotMessage(self, r["time"], r["text"], r["user"], r["room"]))
        
        return resm

if __name__ == "__main__":
    mb = MatrixBot("https://example.com", 'password').SetRoom("xQIjxlkLqVdVKJaxwF")
    mb.Send("Hello World!")


# -*- coding: utf-8 -*-

from .sqlite_platform import SQLitePlatform
from .mysql_platform import MySQLPlatform
from .mysql57_platform import MySQL57Platform
from .postgres_platform import PostgresPlatform



# -*- coding: utf-8 -*-

from .grammar import QueryGrammar
from ...utils import basestring


class MySQLQueryGrammar(QueryGrammar):

    _select_components = [
        "aggregate_",
        "columns",
        "from__",
        "joins",
        "wheres",
        "groups",
        "havings",
        "orders",
        "limit_",
        "offset_",
        "lock_",
    ]

    marker = "%s"

    def compile_select(self, query):
        """
        Compile a select query into SQL

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled sql
        :rtype: str
        """
        sql = super(MySQLQueryGrammar, self).compile_select(query)

        if query.unions:
            sql = "(%s) %s" % (sql, self._compile_unions(query))

        return sql

    def _compile_union(self, union):
        """
        Compile a single union statement

        :param union: The union statement
        :type union: dict

        :return: The compiled union statement
        :rtype: str
        """
        if union["all"]:
            joiner = " UNION ALL "
        else:
            joiner = " UNION "

        return "%s(%s)" % (joiner, union["query"].to_sql())

    def _compile_lock(self, query, value):
        """
        Compile the lock into SQL

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param value: The lock value
        :type value: bool or str

        :return: The compiled lock
        :rtype: str
        """
        if isinstance(value, basestring):
            return value

        if value is True:
            return "FOR UPDATE"
        elif value is False:
            return "LOCK IN SHARE MODE"

    def compile_update(self, query, values):
        """
        Compile an update statement into SQL

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param values: The update values
        :type values: dict

        :return: The compiled update
        :rtype: str
        """
        sql = super(MySQLQueryGrammar, self).compile_update(query, values)

        if query.orders:
            sql += " %s" % self._compile_orders(query, query.orders)

        if query.limit_:
            sql += " %s" % self._compile_limit(query, query.limit_)

        return sql.rstrip()

    def compile_delete(self, query):
        """
        Compile a delete statement into SQL

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :return: The compiled update
        :rtype: str
        """
        table = self.wrap_table(query.from__)

        if isinstance(query.wheres, list):
            wheres = self._compile_wheres(query)
        else:
            wheres = ""

        if query.joins:
            joins = " %s" % self._compile_joins(query, query.joins)

            sql = "DELETE %s FROM %s%s %s" % (table, table, joins, wheres)
        else:
            sql = "DELETE FROM %s %s" % (table, wheres)

        sql = sql.strip()

        if query.orders:
            sql += " %s" % self._compile_orders(query, query.orders)

        if query.limit_:
            sql += " %s" % self._compile_limit(query, query.limit_)

        return sql

    def _wrap_value(self, value):
        """
        Wrap a single string in keyword identifers

        :param value: The value to wrap
        :type value: str

        :return: The wrapped value
        :rtype: str
        """
        if value == "*":
            return value

        return "`%s`" % value.replace("`", "``")



from sklearn.linear_model import SGDClassifier

from .base import * 

# ZH
#                precision    recall  f1-score   support                                                                                                                                                        
#                                                                                                                                                                                                               
#   _08_Finance       0.87      0.94      0.90       453
#        _10_IT       0.87      0.89      0.88       476
#    _13_Health       0.92      0.91      0.92       496
#    _14_Sports       0.99      0.99      0.99       490
#    _16_Travel       0.93      0.92      0.93       496
# _20_Education       0.88      0.91      0.90       474
#   _22_Recruit       0.93      0.89      0.91       514
#   _23_Culture       0.88      0.85      0.87       511
#  _24_Military       0.97      0.95      0.96       500
# 
#      accuracy                           0.92      4410
#     macro avg       0.92      0.92      0.92      4410
#  weighted avg       0.92      0.92      0.92      4410
# 
# EN
#                           precision    recall  f1-score   support                                                                                                                                             
#                                                                                                                                                                                                               
#              alt.atheism       0.77      0.82      0.79       299
#            comp.graphics       0.78      0.76      0.77       401
#  comp.os.ms-windows.misc       0.76      0.76      0.76       394
# comp.sys.ibm.pc.hardware       0.76      0.72      0.74       410
#    comp.sys.mac.hardware       0.85      0.83      0.84       391
#           comp.windows.x       0.76      0.87      0.81       344
#             misc.forsale       0.91      0.82      0.86       434
#                rec.autos       0.88      0.93      0.91       377
#          rec.motorcycles       0.95      0.95      0.95       399
#       rec.sport.baseball       0.95      0.91      0.93       416
#         rec.sport.hockey       0.98      0.95      0.96       413
#                sci.crypt       0.95      0.93      0.94       407
#          sci.electronics       0.74      0.80      0.77       367
#                  sci.med       0.89      0.91      0.90       390
#                sci.space       0.95      0.89      0.92       418
#   soc.religion.christian       0.93      0.84      0.89       442
#       talk.politics.guns       0.94      0.74      0.83       458
#    talk.politics.mideast       0.90      0.97      0.94       349
#       talk.politics.misc       0.61      0.85      0.71       220
#       talk.religion.misc       0.61      0.75      0.67       203
# 
#                 accuracy                           0.85      7532
#                macro avg       0.84      0.85      0.84      7532
#             weighted avg       0.86      0.85      0.85      7532

class SVM(baseClassificationClass):
    def __init__(self) -> None:
        super().__init__()
        self.classifierfunc = SGDClassifier

if __name__ == "__main__":
    from bagbag import * 

    traindatadir = Os.Getenv("HOME") + '/data/train'
    testdatadir = Os.Getenv("HOME") + '/data/test'

    tc = Tools.TextClassifier.SVM()

    # # ipdb.set_trace()

    for label in Tools.ProgressBar(Os.ListDir(traindatadir)):
        fdir = Os.Path.Join(traindatadir, label)
        for fname in Tools.ProgressBar(Os.ListDir(fdir)):
            fpath = Os.Path.Join(fdir, fname)
            text = open(fpath, encoding='gbk', errors='ignore').read()
            tc.Add(label, text)

    Lg.Trace("training")
    tc.Train()

    tc.Save("tc.obj")

    del(tc)

    ts = Tools.TextClassifier.SVM()
    ts.Load("tc.obj")

    ay = []
    py = []
    for label in Tools.ProgressBar(Os.ListDir(testdatadir)):
        fdir = Os.Path.Join(testdatadir, label)
        for fname in Tools.ProgressBar(Os.ListDir(fdir)):
            fpath = Os.Path.Join(fdir, fname)
            text = open(fpath, encoding='gbk', errors='ignore').read()
            # tc.Add(label, text)

            ay.append(label)
            py.append(ts.Predict(text))

    print(ts.Report(py, ay))
        




import subprocess
import typing 
import collections
import queue
import threading
import json
import subprocess
import shutil

# print("load cmd")

class ContinuousSubprocess:
    """
    Creates a process to execute a wanted command and
    yields a continuous output stream for consumption.
    """

    def __init__(self, command_string: str) -> None:
        """
        Constructor.

        :param command_string: A command to execute in a separate process.
        """
        self.__command_string = command_string
        self.__process: typing.Optional[subprocess.Popen] = None
        self.terminated = False

    @property
    def command_string(self) -> str:
        """
        Property for command string.

        :return: Command string.
        """
        return self.__command_string

    def Terminate(self) -> None:
        if not self.__process:
            raise ValueError('Process is not running.')

        # self.__process.terminate()
        self.__process.kill()
        self.terminated = True

    def Execute(
        self,
        shell: bool = True,
        path: typing.Optional[str] = None,
        max_error_trace_lines: int = 1000,
        *args,
        **kwargs,
    ) -> typing.Generator[str, None, None]:
        """
        Executes a command and yields a continuous output from the process.

        :param shell: Boolean value to specify whether to
        execute command in a new shell.
        :param path: Path where the command should be executed.
        :param max_error_trace_lines: Maximum lines to return in case of an error.
        :param args: Other arguments.
        :param kwargs: Other named arguments.

        :return: A generator which yields output strings from an opened process.
        """
        # Check if the process is already running (if it's set, then it means it is running).
        if self.__process:
            raise RuntimeError(
                'Process is already running. '
                'To run multiple processes initialize a second object.'
            )

        with subprocess.Popen(
            self.__command_string,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
            shell=shell,
            cwd=path,
            *args,
            **kwargs,
        ) as process:

            # Indicate that the process has started and is now running.
            self.__process = process

            # Initialize a mutual queue that will hold stdout and stderr messages.
            q = queue.Queue()
            # Initialize a limited queue to hold last N of lines.
            dq = collections.deque(maxlen=max_error_trace_lines)

            # Create a parallel thread that will read stdout stream.
            stdout_thread = threading.Thread(
                target=ContinuousSubprocess.__read_stream, args=[process.stdout, q]
            )
            stdout_thread.start()

            # Create a parallel thread that will read stderr stream.
            stderr_thread = threading.Thread(
                target=ContinuousSubprocess.__read_stream, args=[process.stderr, q]
            )
            stderr_thread.start()

            # Run this block as long as our main process is alive or std streams queue is not empty.
            while (process.poll() is None) or (not q.empty()):
                try:
                    # Rad messages produced by stdout and stderr threads.
                    item = q.get(block=True, timeout=1)
                    dq.append(item)
                    yield item
                    if self.terminated == True:
                        return 
                except queue.Empty:
                    pass

            # Close streams.
            process.stdout.close()
            process.stderr.close()

            return_code = process.wait()

        # Make sure both threads have finished.
        stdout_thread.join(timeout=5)
        if stdout_thread.is_alive():
            raise RuntimeError('Stdout thread is still alive!')

        stderr_thread.join(timeout=5)
        if stderr_thread.is_alive():
            raise RuntimeError('Stderr thread is still alive!')

        # Indicate that the process has finished as is no longer running.
        self.__process = None

        if return_code:
            error_trace = list(dq)
            raise subprocess.CalledProcessError(
                returncode=return_code,
                cmd=self.__command_string,
                output=json.dumps(
                    {
                        'message': 'An error has occurred while running the specified command.',
                        'trace': error_trace,
                        'trace_size': len(error_trace),
                        'max_trace_size': max_error_trace_lines,
                    }
                ),
            )

    @staticmethod
    def __read_stream(stream: typing.IO[typing.AnyStr], queue: queue.Queue):
        try:
            for line in iter(stream.readline, ''):
                if line != '':
                    queue.put(line)
        # It is possible to receive: ValueError: I/O operation on closed file.
        except ValueError:
            return

# command = Os.Args[1]
# generator = ContinuousSubprocess(command).execute()

# for data in generator:
#     print(data)

def TailOutput(cmd:str) -> typing.Iterator[str]:
    for i in ContinuousSubprocess(cmd).Execute():
        if i[-1] == "\n":
            yield i[:-1]
        else:
            yield i

def GetStatusOutput(cmd:str, decodeerror:str="ignore") -> typing.Tuple[int, str]:
    """
    函数 `GetStatusOutput` 执行一个 shell 命令，并返回一个包含状态码和命令输出的元组。

    参数：
    - `cmd`：一个字符串，表示要在 shell 中执行的命令。它将被传递给 `subprocess.Popen` 函数，以在子进程中运行该命令。
    - `decodeerror`：在 `GetStatusOutput` 函数中，`decodeerror` 参数用于指定在将命令输出转换为字符串时如何处理解码错误。默认设置为 "ignore"，这意味着在解码过程中遇到的任何解码错误都将被忽略。可选strict

    返回：一个元组，包含一个表示命令执行状态的整数和一个表示命令输出的字符串。
    """
    tsk = subprocess.Popen(["sh", "-c", cmd],stdout=subprocess.PIPE,stderr=subprocess.STDOUT)
    status = tsk.wait()

    return status, tsk.stdout.read().decode('utf-8', decodeerror)

def GetOutput(cmd:str, decodeerror:str='ignore') -> str:
    """
    `decodeerror` 参数用于指定在将命令输出转换为字符串时如何处理解码错误。默认设置为 "ignore"，这意味着在解码过程中遇到的任何解码错误都将被忽略。可选strict
    """
    _, output = GetStatusOutput(cmd, decodeerror)

    return output

def Exist(cmd:str) -> bool:
    return shutil.which(cmd) is not None

def Where(cmd:str) -> str:
    return shutil.which(cmd)

if __name__ == "__main__":
    print(Exist("ls"))


# -*- coding: utf-8 -*-

import simplejson as json
from wrapt import ObjectProxy
from ..utils import value


class Dynamic(ObjectProxy):

    _key = None
    _fluent = None

    def __init__(self, value, key, fluent):
        super(Dynamic, self).__init__(value)

        self._key = key
        self._fluent = fluent

    def __call__(self, *args, **kwargs):
        if len(args):
            self.__set_value(args[0])
        else:
            self.__set_value(True)

        return self._fluent

    def __set_value(self, value):
        self._fluent._attributes[self._key] = value


class Fluent(object):
    def __init__(self, **attributes):
        self._attributes = {}

        for key, value in attributes.items():
            self._attributes[key] = value

    def get(self, key, default=None):
        return self._attributes.get(key, value(default))

    def get_attributes(self):
        return self._attributes

    def to_dict(self):
        return self.serialize()

    def serialize(self):
        return self._attributes

    def to_json(self, **options):
        return json.dumps(self.serialize(), **options)

    def __contains__(self, item):
        return item in self._attributes

    def __getitem__(self, item):
        return self._attributes[item]

    def __setitem__(self, key, value):
        self._attributes[key] = value

    def __delitem__(self, key):
        del self._attributes[key]

    def __dynamic(self, method):
        def call(*args, **kwargs):
            if len(args):
                self._attributes[method] = args[0]
            else:
                self._attributes[method] = True

            return self

        return call

    def __getattr__(self, item):
        return Dynamic(self._attributes.get(item), item, self)

    def __setattr__(self, key, value):
        if key == "_attributes":
            super(Fluent, self).__setattr__(key, value)

        try:
            super(Fluent, self).__getattribute__(key)

            return super(Fluent, self).__setattr__(key, value)
        except AttributeError:
            pass

        self._attributes[key] = value

    def __delattr__(self, item):
        del self._attributes[item]



# -*- coding: utf-8 -*-

from .processor import QueryProcessor


class PostgresQueryProcessor(QueryProcessor):
    def process_insert_get_id(self, query, sql, values, sequence=None):
        """
        Process an "insert get ID" query.

        :param query: A QueryBuilder instance
        :type query: QueryBuilder

        :param sql: The sql query to execute
        :type sql: str

        :param values: The value bindings
        :type values: list

        :param sequence: The ids sequence
        :type sequence: str

        :return: The inserted row id
        :rtype: int
        """
        result = query.get_connection().select_from_write_connection(sql, values)

        id = result[0][0]

        if isinstance(id, int):
            return id

        if str(id).isdigit():
            return int(id)

        return id

    def process_column_listing(self, results):
        """
        Process the results of a column listing query

        :param results: The query results
        :type results: dict

        :return: The processed results
        :return: list
        """
        return list(map(lambda x: x["column_name"], results))



import time 

#print("load " + '/'.join(__file__.split('/')[-2:]))

pformat = (lambda a:lambda v,t="    ",n="\n",i=0:a(a,v,t,n,i))(lambda f,v,t,n,i:"{%s%s%s}"%(",".join(["%s%s%s: %s"%(n,t*(i+1),repr(k),f(f,v[k],t,n,i+1))for k in v]),n,(t*i)) if type(v)in[dict] else (type(v)in[list]and"[%s%s%s]"or"(%s%s%s)")%(",".join(["%s%s%s"%(n,t*(i+1),f(f,k,t,n,i+1))for k in v]),n,(t*i)) if type(v)in[list,tuple] else repr(v))

def Size(ByteNumber, suffix='B'):
    for unit in ['','K','M','G','T','P','E','Z']:
        if abs(ByteNumber) < 1024.0:
            return "%3.1f%s%s" % (ByteNumber, unit, suffix)
        ByteNumber /= 1024.0
    return "%.1f%s%s" % (ByteNumber, 'Y', suffix)

def TimeDuration(seconds:int) -> str:
    return time.strftime("%H:%M:%S", time.gmtime(seconds))

def PrettyJson(obj:dict|list) -> str:
    return pformat(obj)


import cachetools

#print("load " + '/'.join(__file__.split('/')[-2:]))

def LRU(size:int) -> cachetools.LRUCache:
    """
    The function returns an instance of a Least Recently Used (LRU) cache with a specified size using
    the cachetools library in Python.
    
    :param size: The parameter "size" is an integer that specifies the maximum number of items that can
    be stored in the LRU cache. When the cache reaches its maximum capacity, the least recently used
    item will be removed to make space for a new item
    :type size: int
    :return: The function `LRU` returns an instance of the `cachetools.LRUCache` class with the
    specified `size` parameter.
    """
    return cachetools.LRUCache(size)

def FIFO(size:int) -> cachetools.FIFOCache:
    """
    This function returns a FIFO cache object with a specified size using the cachetools library in
    Python.
    
    :param size: The parameter "size" is an integer that specifies the maximum number of items that can
    be stored in the FIFO cache
    :type size: int
    :return: an instance of the `cachetools.FIFOCache` class with the specified `size` parameter.
    """
    return cachetools.FIFOCache(size)

def LFU(size:int) -> cachetools.LFUCache:
    """
    The function returns an instance of a Least Frequently Used (LFU) cache with a specified size using
    the cachetools library in Python.
    
    :param size: The parameter "size" is an integer that specifies the maximum number of items that can
    be stored in the LFU cache. The LFU cache is a type of cache that stores items based on their
    frequency of use, with the least frequently used items being evicted first when the cache reaches
    its maximum
    :type size: int
    :return: The function `LFU` returns an instance of the `cachetools.LFUCache` class with the
    specified `size` parameter.
    """
    return cachetools.LFUCache(size)

def MRU(size:int) -> cachetools.MRUCache:
    """
    The function creates and returns a MRU cache object with a specified size using the cachetools
    library in Python.
    
    :param size: The parameter "size" is an integer that represents the maximum number of items that can
    be stored in the MRU cache
    :type size: int
    :return: The function `MRU` returns an instance of the `cachetools.MRUCache` class with the
    specified `size` parameter.
    """
    return cachetools.MRUCache(size)

def RR(size:int) -> cachetools.RRCache:
    """
    The function creates and returns a cache object with a specified size using the RRCache algorithm
    from the cachetools library in Python.
    
    :param size: The parameter "size" is an integer that specifies the maximum number of items that can
    be stored in the RRCache object. The RRCache is a cache implementation that uses a "recently read"
    eviction policy, which means that the least recently read items will be evicted from the cache
    :type size: int
    :return: The function `RR` returns an instance of the `cachetools.RRCache` class with the specified
    `size` parameter.
    """
    return cachetools.RRCache(size)

def TTL(size:int) -> cachetools.TTLCache:
    """
    The function creates and returns a TTLCache object with a specified size using the cachetools
    library.
    
    :param size: The parameter "size" is an integer that specifies the maximum number of items that can
    be stored in the cache. It is used to initialize a new instance of the TTLCache class from the
    cachetools module. This cache implementation automatically removes items that have not been accessed
    for a certain amount of time
    :type size: int
    :return: The function `TTL` is returning an instance of `cachetools.TTLCache` with the specified
    `size` parameter.
    """
    return cachetools.TTLCache(size)

if __name__ == "__main__":
    # from bagbag import Range, Funcs, Time, Lg
    cache = LRU(5000)

    dic = {}
    keys = []
    for _ in Range(5000):
        key = Funcs.UUID()
        cache[key] = Funcs.UUID()
        keys.append(key)
        dic[key] = None 

    start = Time.Now()

    for _ in Range(10000):
        # cache[Random.Choice(keys)] # 0.006376028060913086
        # if Funcs.UUID() in cache: #  0.038994789123535156
        # if Random.Choice(keys) in cache: # 0.004488945007324219
        # try:
        #     # cache[Random.Choice(keys)] # 0.006552934646606445
        #     cache[Funcs.UUID()] # 0.04251289367675781
        # except:
        #     pass 
        # cache.__contains__(Funcs.UUID()) # 0.03940105438232422
        Funcs.UUID() in dic

    end = Time.Now()

    Lg.Trace(end - start)


from .... import Http, Json
from ... import RateLimit

class API():
    def __init__(self, key:str) -> None:
        self.key = key 
        self.headers = {'Ok-Access-Key': self.key}
        self.domain = 'www.oklink.com'
        self.rl = RateLimit("5/s")
    
    def CheckLabel(self, chan:str, address:list[str]) -> list[dict]:
        '''
        只可以批量查询20个同一个链的地址的标签. 支持的链有：BTC, BCH, LTC, DASH, DOGE, ETH, OKTC, XLAYER, BSC, ETC, POLYGON, AVAXC, ETHW, DIS, FTM, OP, ARBITRUM, KLAYTN, ZKSYNC, GNOSIS, RONIN, LINEA, POLYGON_ZKEVM, APT, SUI, TRON, STARKNET, BASE, SCROLL, OMEGA, OPBNB
        返回的数据格式是
        [
            {
                "label": [
                    "OKX.Cold Wallet"
                ],
                "address": "0x539c92186f7c6cc4cbf443f26ef84c595babbca1"
            }
        ]
        '''

        address = ','.join(address)
        url = f'https://{self.domain}/api/v5/explorer/address/entity-label?chainShortName={chan}&address={address}'

        self.rl.Take()
        resp = Http.Get(url, headers=self.headers)

        if resp.StatusCode != 200:
            raise Exception(f"状态码不为200: {resp.StatusCode} ==> {resp.Content}")
        
        content = Json.Loads(resp.Content)
        data = content['data']

        resp = []
        for item in data:
            resp.append({
                'label': item['label'].split(','),
                "address": item['address']
            })

        return resp


import zipfile
import os
from collections import defaultdict
import typing

class ZIP():
    def __init__(self, zip_file_or_dir:str) -> None:
        self.path = zip_file_or_dir

        self.zipfr:zipfile.ZipFile = None

    def open_for_read(self):
        if self.zipfr == None:
            self.zipfr = zipfile.ZipFile(self.path, 'r')
    
    def Pack(self, out_zip_fpath:str):
        """
        将指定目录中的所有文件和子目录打包为一个zip文件

        :param input_dir: 要打包的目录路径
        :param output_zip_file: 输出的zip文件路径
        """
        with zipfile.ZipFile(out_zip_fpath, 'w', zipfile.ZIP_DEFLATED) as zipf:
            self.zipfr = None 
            for root, dirs, files in os.walk(self.path):
                for file in files:
                    file_path = os.path.join(root, file)
                    # Arcname should be relative to the input_dir
                    arcname = os.path.relpath(file_path, start=self.path)
                    zipf.write(file_path, arcname)

    def Unpack(self, extract_to_dir:str):
        """
        The function `Unpack` extracts all files from a zip archive to a specified directory after
        ensuring the directory exists.
        
        :param extract_to_dir: The `extract_to_dir` parameter is a string that represents the directory
        path where you want to extract the contents of the zip file. This function will check if the
        directory exists, create it if it doesn't, and then extract all files from the zip file to that
        directory
        :type extract_to_dir: str
        """
        # 确保解压目录存在
        if not os.path.exists(extract_to_dir):
            os.makedirs(extract_to_dir)
        
        self.open_for_read()
        
        self.zipfr.extractall(extract_to_dir)

    def Read(self, fpath:str) -> str:
        """
        The function `Read` reads and returns the contents of a file within a zip archive specified by
        `fpath`.
        
        :param fpath: The `fpath` parameter in the `Read` method is a string that represents the file
        path within the zip archive from which you want to read the contents
        :type fpath: str
        :return: The `Read` method is returning the content of a file located at the specified `fpath`
        within a zip archive. The content is read as bytes and then decoded into a UTF-8 encoded string
        before being returned.
        """
        self.open_for_read()

        with self.zipfr.open(fpath) as file:
            return file.read().decode('utf-8')
    
    def ReadBytes(self, fpath:str) -> bytes:
        """
        The function `ReadBytes` reads and returns the contents of a file within a zip archive as bytes.
        
        :param fpath: The `fpath` parameter in the `ReadBytes` method is a string that represents the
        file path of the file you want to read from the zip archive. When calling the `ReadBytes`
        method, you should provide the specific file path within the zip archive that you want to read
        as a
        :type fpath: str
        :return: The `ReadBytes` method reads and returns the contents of a file located at the
        specified `fpath` within a zip archive. It returns the contents of the file as a bytes object.
        """
        self.open_for_read()

        with self.zipfr.open(fpath) as file:
            return file.read()
    
    def zip_os_walk(self):
        def split_path(path):
            """Split a path into its segments."""
            parts = []
            while path:
                path, tail = os.path.split(path)
                if tail:
                    parts.insert(0, tail)
                else:
                    if path:
                        parts.insert(0, path)
                    break
            return parts

        self.open_for_read()
        directories = defaultdict(lambda: {'dirs': [], 'files': []})

        # Collect directory and file information
        for zip_info in self.zipfr.infolist():
            parts = split_path(zip_info.filename)
            parent_dir = os.path.join(*parts[:-1]) if parts[:-1] else ''
            if zip_info.is_dir():
                directories[parent_dir]['dirs'].append(parts[-1])
            else:
                directories[parent_dir]['files'].append(parts[-1])

        # Yield directory content similar to os.walk
        for dirpath in directories:
            yield dirpath, directories[dirpath]['dirs'], directories[dirpath]['files']

    def Walk(self, type:str=None) -> typing.Iterable[typing.Tuple[str, str, str]]:
        for root, dirs, files in self.zip_os_walk():
            if type == None:
                for name in files:
                    yield os.path.join(root, name)
                for name in dirs:
                    yield os.path.join(root, name)
            elif type == "f":
                for name in files:
                    yield os.path.join(root, name)
            elif type == "d":
                for name in dirs:
                    yield os.path.join(root, name)
    
    def Close(self):
        if self.zipfr != None:
            self.zipfr.close()


from pythonping import ping

from ..Tools import Chan
from ..Thread import Thread
from ..String import String

#print("load " + __file__.split('/')[-1])

class filelike():
    def __init__(self, c):
        self.c = c 

    def write(self, msg):
        msg = msg.strip()
        if msg != "":
            if 'timed out' in msg:
                self.c.Put("timeout")
            else:
                self.c.Put(float(String(msg).RegexFind("Reply from .+?, .+? bytes in (.+)ms")[0][0]))

def Ping(host, timeout:int=3, count:int=None, interval:int=1):
    c = Chan(0)
    fd = filelike(c)
    def run():
        if count:
            ping(host, timeout=timeout, count=count, interval=interval, verbose=True, out=fd)
        else:
            while True:
                ping(host, timeout=timeout, count=60, interval=interval, verbose=True, out=fd)
        c.Close()
    Thread(run)
    return c

if __name__ == "__main__":
    while True:
        for i in Ping("8.8.8.8"):
            Lg.Trace(i)



try:
    from .Lock_src import Lock
except:
    from Lock_src import Lock

import time

#print("load " + '/'.join(__file__.split('/')[-2:]))

class WaitGroup():
    def __init__(self):
        self.count = 0
        self.lock = Lock()
    
    def Add(self):
        self.lock.Acquire()
        self.count += 1
        self.lock.Release()
    
    def Done(self):
        self.lock.Acquire()
        self.count -= 1
        self.lock.Release()
    
    def Wait(self, timeout:int=-1) -> bool:
        waitedsec = 0 
        while True:
            if self.count == 0:
                return True
            else:
                if timeout != -1 and waitedsec >= timeout:
                    return False 
                time.sleep(1)
                waitedsec += 1


# -*- coding: utf-8 -*-

from .pivot import Pivot


class MorphPivot(Pivot):

    _morph_name = None
    _morph_type = None

    def _set_keys_for_save_query(self, query):
        """
        Set the keys for a save update query.

        :param query: A Builder instance
        :type query: orator.orm.Builder

        :return: The Builder instance
        :rtype: orator.orm.Builder
        """
        query.where(self._morph_type, self._morph_name)

        return super(MorphPivot, self)._set_keys_for_save_query(query)

    def delete(self):
        """
        Delete the pivot model record from the database.

        :rtype: int
        """
        query = self._get_delete_query()

        query.where(self._morph_type, self._morph_name)

        return query.delete()

    def set_morph_type(self, morph_type):
        self._morph_type = morph_type

        return self

    def set_morph_name(self, morph_name):
        self._morph_name = morph_name

        return self




from typing import Any


#print("load random")

def Int(min:int, max:int) -> int:
    import random 
    return random.randint(min, max)

def Choice(obj:list|str, count:int=1) -> Any | list[Any]:
    import random 
    if count <= 1:
        return random.choice(obj)
    else:
        return [random.choice(obj) for i in range(count)]

def String(length:int=8, charset:str="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789") -> str:
    import random 
    res = []
    while len(res) < length:
        res.append(random.choice(charset))
    
    return "".join(res)

def Shuffle(li:list) -> list:
    import random 
    import copy
    l = copy.copy(li)
    random.shuffle(l)
    return l

if __name__ == "__main__":
    print(Choice("doijwoefwe"))
    print(String(5))
    print(Shuffle([1,2,3,4,5]))


# -*- coding: utf-8 -*-

__version__ = "0.9.9"

from .orm import Model, SoftDeletes, Collection, accessor, mutator, scope
from .database_manager import DatabaseManager
from .query.expression import QueryExpression
from .schema import Schema
from .pagination import Paginator, LengthAwarePaginator



# 配合以下docker-compose使用
# 
# version: '3'
# services:
#   ocr-server:
#     image: darren2046/ocr-server
#     networks:
#       ocr-server-vpc:
#         ipv4_address: 192.168.168.63
#     container_name: ocr-server-192.168.168.63
#     restart: always
#     ports:
#       - 8990:8990
# 

#print("load " + '/'.join(__file__.split('/')[-2:]))

from .. import Http
from .. import Base64
from .. import Json

class ocrResultText():
    def __init__(self, Coordinate:list, Text:str, Confidence:float) -> None:
        self.Coordinate:list = Coordinate 
        self.Text:str = Text 
        self.Confidence:float = Confidence 

class ocrResult():
    def __init__(self) -> None:
        self.Data:bytes = b'' 
        self.Texts:list[ocrResultText] = [] 
    
    def SaveImage(self, fpath:str):
        """
        保存JPEG内容到文件路径, 覆盖已有文件.
        
        :param fpath: fpath is a string parameter that represents the file path where the image will be
        saved. It is used as an argument for the open() function to create a file object in binary write
        mode ('wb'). The image data is then written to this file object using the write() method, and
        the file
        :type fpath: str
        """
        fd = open(fpath, 'wb')
        fd.write(self.Data)
        fd.close()

class OCR():
    def __init__(self, server:str) -> None:
        self.server = server

        if not self.server.startswith('http://') and not self.server.startswith("https://"):
            self.server = 'https://' + self.server

    def Recognition(self, fpath:str, lang:str="ch") -> ocrResult:
        data = open(fpath, 'rb').read()

        resp = Http.PostJson(self.server + "/ocr", {
            "lang": lang, 
            "data": Base64.Encode(data),
        })

        r = Json.Loads(resp.Content)

        if r['code'] != 200:
            raise Exception("识别OCR出错:" + r['message'])
        
        ocrr = ocrResult()
        ocrr.Data = Base64.Decode(r['data'])
        for rr in r['result']:
            ocrr.Texts.append(ocrResultText(rr['coordinate'], rr['text'], rr['confidence']))

        return ocrr 
    
if __name__ == "__main__":
    ocr = OCR("api.svc.ltd")

    result = ocr.Recognition("foo.png")

    result.SaveImage("foo.result.jpg")

    for r in result.Texts:
        print(r.Text, r.Confidence)


from __future__ import annotations

from selenium import webdriver
from selenium.webdriver.common.by import By as webby
import selenium
from selenium.webdriver.common.keys import Keys as webkeys
from selenium.webdriver.firefox.options import Options as firefoxoptions
from selenium.webdriver.chrome.options import Options as chromeoptions

# from seleniumwire import webdriver as seleniumwirewebdriver
# from seleniumwire.utils import decode as decodeResponseBody

import time
import random
import typing
import copy
import base64

#print("load " + '/'.join(__file__.split('/')[-2:]))

from ..Http import useragents 
from .. import Lg
from ..Thread import Thread
from .URL_src import URL
from ..String import String 
from .. import Os
from ..File import File
from .. import Cmd

seleniumChromeWireSkipFilesSuffix = (
    # 图片
    '.png', 
    '.jpg', 
    '.gif', 
    '.jpeg', 
    '.tiff', 
    '.psd', 
    '.raw', 
    '.webp', 
    '.eps',
    '.svg',
    '.bmp',
    '.pdf',
    '.pcx',
    '.tga',
    '.exif',
    '.fpx',
    # 视频
    '.avi',
    '.wmv',
    '.mpg',
    '.mpeg',
    '.mov',
    '.rm',
    '.rmvb',
    '.swf',
    '.flv',
    '.mp4',
    '.asf',
    '.dat',
    '.asx',
    '.wvx',
    '.mpe',
    '.mpa',
    # 音频
    '.mp3',
    '.wma',
    '.wav',
    '.mid',
    '.ape',
    '.flac',

    # 其它
    # '.aiff', # 声音文件
    # '.aej', '.cab', '.rar', # 压缩文件
    # '.awd', # 传真文件
    # '.bak', #备份文件
    # '.scr', #屏保文件
    # '.sys', #系统文件
    # '.ttf', '.font', #字体文件
    # '.doc', #文档文件
)
    
def retryOnError(func):
    def ware(self, *args, **kwargs): # self是类的实例
        if self.ToRetryOnError == False:
            res = func(self, *args, **kwargs)
            return  
        
        if self.browserName in ["chrome", "chromewire"]:
            while True:
                try:
                    res = func(self, *args, **kwargs)

                    try:
                        NeedRefresh = False
                        # 如果载入页面失败, 有个Reload的按钮
                        if hasattr(self, "Find"):
                            if self._find("/html/body/div[1]/div[2]/div/button[1]", 0):
                                if self._find("/html/body/div[1]/div[2]/div/button[1]").Text() == "Reload":
                                    NeedRefresh = True
                        elif hasattr(self, "se") and hasattr(self.se, "Find"):
                            if self.se._find("/html/body/div[1]/div[2]/div/button[1]", 0):
                                if self.se._find("/html/body/div[1]/div[2]/div/button[1]").Text() == "Reload":
                                    NeedRefresh = True

                        if hasattr(self, "PageSource"):
                            page = self.PageSource()
                        elif hasattr(self, "se") and hasattr(self.se, "PageSource"):
                            page = self.se.PageSource()
                        
                        if hasattr(self, "Url"):
                            url = self.Url()
                        elif hasattr(self, "se") and hasattr(self.se, "Url"):
                            url = self.se.Url()

                        chklists = [
                            [
                                'This page isn’t working',
                                'ERR_EMPTY_RESPONSE',
                                'didn’t send any data',
                                URL(url).Parse().Host,
                            ],
                            [
                                'This site can’t be reached',
                                'unexpectedly closed the connection',
                                'ERR_CONNECTION_CLOSED',
                                URL(url).Parse().Host,
                            ],
                            [
                                "This site can’t be reached",
                                "took too long to respond",
                                "ERR_TIMED_OUT",
                                URL(url).Parse().Host,
                            ],
                            [
                                "No internet",
                                "There is something wrong with the proxy server, or the address is incorrect",
                                "ERR_PROXY_CONNECTION_FAILED",
                            ],
                            [
                                'ERR_CONNECTION_RESET',
                                'This site can’t be reached',
                                'The connection was reset',
                                URL(url).Parse().Host,
                            ]
                        ]
                        for chklist in chklists:
                            if False not in map(lambda x: x in page, chklist):
                                NeedRefresh = True 
                        
                        if NeedRefresh:
                            if hasattr(self, "Refresh"):
                                self.Refresh()
                            elif hasattr(self, "se") and hasattr(self.se, "Refresh"):
                                self.se.Refresh()
                            time.sleep(5)
                        else:
                            return res

                    except Exception as e:
                        if hasattr(self, "closed") and self.closed:
                            break 
                        elif hasattr(self, "se") and hasattr(self.se, "closed") and self.se.closed:
                            break
                        else:
                            raise e
                    time.sleep(1)
                except Exception as e:
                    chklist = [
                        'ERR_CONNECTION_CLOSED',
                        'ERR_EMPTY_RESPONSE',
                        'ERR_TIMED_OUT',
                        'ERR_PROXY_CONNECTION_FAILED',
                        'ERR_CONNECTION_RESET',
                    ]
                    if True in map(lambda x: x in str(e), chklist):
                        Lg.Trace("有错误, 自动刷新")
                        if hasattr(self, "Refresh"):
                            self.Refresh()
                        elif hasattr(self, "se") and hasattr(self.se, "Refresh"):
                            self.se.Refresh()
                        time.sleep(5)
                    else:
                        raise e

        elif self.browserName == "firefox":
            res = func(self, *args, **kwargs)
            return 

    return ware

# > The SeleniumElement class is a wrapper for the selenium.webdriver.remote.webelement.WebElement
# class
class SeleniumElement():
    def __init__(self, element:selenium.webdriver.remote.webelement.WebElement, se:seleniumBase, ToRetryOnError:bool):
        self.element = element
        self.se = se
        self.driver = self.se.driver
        self.browserName = self.se.browserName
        self.browserRemote = self.se.browserRemote
        self.ToRetryOnError = ToRetryOnError
    
    def Clear(self) -> SeleniumElement:
        """
        Clear() clears the text if it's a text entry element
        """
        self.element.clear()
        return self
    
    @retryOnError
    def Click(self) -> SeleniumElement:
        """
        Click() is a function that clicks on an element
        """
        if self.browserName in ["chrome", "chromewire"] and not self.browserName:
            self.driver.execute_cdp_cmd('Network.setUserAgentOverride', {"userAgent": random.choice(useragents)['user_agent']})

        self.element.click()

        return self
    
    def Text(self) -> str:
        """
        The function Text() returns the text of the element
        :return: The text of the element.
        """
        return self.element.text

    def Attribute(self, name:str) -> str:
        """
        This function returns the value of the attribute of the element
        
        :param name: The name of the element
        :type name: str
        :return: The attribute of the element.
        """
        return self.element.get_attribute(name)
    
    def Input(self, string:str) -> SeleniumElement:
        """
        The function Input() takes in a string and sends it to the element
        
        :param string: The string you want to input into the text box
        :type string: str
        """
        self.element.send_keys(string)
        return self
    
    @retryOnError
    def Submit(self) -> SeleniumElement:
        """
        Submit() is a function that submits the form that the element belongs to
        """
        if self.browserName in ["chrome", "chromewire"] and not self.browserName:
            self.driver.execute_cdp_cmd('Network.setUserAgentOverride', {"userAgent": random.choice(useragents)['user_agent']})
        
        self.element.submit()

        return self
    
    @retryOnError
    def PressEnter(self) -> SeleniumElement:
        """
        It takes the element that you want to press enter on and sends the enter key to it
        """

        if self.browserName  in ["chrome", "chromewire"] and not self.browserName:
            self.driver.execute_cdp_cmd('Network.setUserAgentOverride', {"userAgent": random.choice(useragents)['user_agent']})
        
        self.element.send_keys(webkeys.ENTER)

        return self
    
    def ScrollIntoElement(self) -> SeleniumElement:
        self.driver.execute_script("arguments[0].scrollIntoView(true);", self.element)
        return self

    def HTML(self) -> str:
        return self.element.get_attribute('innerHTML')
    
    def Image(self) -> bytes:
        # 执行JavaScript将图片转换为Base64编码
        captcha_base64 = self.driver.execute_script("""
            var img = arguments[0];
            var canvas = document.createElement('canvas');
            canvas.width = img.width;
            canvas.height = img.height;
            var ctx = canvas.getContext('2d');
            ctx.drawImage(img, 0, 0, img.width, img.height);
            return canvas.toDataURL('image/png').split(',')[1];
        """, self.element)

        return base64.b64decode(captcha_base64)

class seleniumBase():
    def Find(self, xpath:str|list, timeout:int=60, scrollIntoElement:bool=True, retryOnError:int=5) -> SeleniumElement|None:
        """
        This is a Python function that finds a Selenium element based on an XPath string or list of XPath
        strings, with optional parameters for timeout, scrolling, and retrying on errors.
        
        :param xpath: The xpath expression used to locate the element(s) on the webpage
        :type xpath: str|list
        :param timeout: The maximum amount of time (in seconds) to wait for the element to be found before
        timing out and returning None, defaults to 60
        :type timeout: int (optional)
        :param scrollIntoElement: scrollIntoElement is a boolean parameter that determines whether the web
        page should be scrolled to bring the element into view before attempting to find it using the
        specified XPath. If set to True, the page will be scrolled to bring the element into view before
        attempting to find it. If set to False, the, defaults to True
        :type scrollIntoElement: bool (optional)
        :param retryOnError: The retryOnError parameter specifies the number of times the function should
        retry finding the element if an error occurs during the search. If the element is not found after
        the specified number of retries, the function will return None, defaults to 5
        :type retryOnError: int (optional)
        :return: The function `Find` returns a `SeleniumElement` object if it is found, or `None` if it is
        not found.
        """
        if type(xpath) == str:
            return self._find(xpath, timeout, scrollIntoElement, retryOnError)
        else:
            idx = self.Except(xpath, timeout=timeout)
            if idx == None:
                return None 
            else:
                return self._find(xpath[idx], timeout, scrollIntoElement, retryOnError)

    def _find(self, xpath:str, timeout:int=60, scrollIntoElement:bool=True, retryOnError:int=5) -> SeleniumElement|None:
        """
        > Finds an element by xpath, waits for it to appear, and returns it
        
        :param xpath: The xpath of the element you want to find
        :type xpath: str
        :param timeout: , defaults to 8 second
        :type timeout: int (optional)
        :param scrollIntoElement: If True, the element will be scrolled into view before returning it,
        defaults to True
        :type scrollIntoElement: bool (optional)
        :return: SeleniumElement
        """
        waited = 0
        errcount = 0
        while True:
            try:
                el = self.driver.find_element(webby.XPATH, xpath)
                if scrollIntoElement:
                    self.driver.execute_script("arguments[0].scrollIntoView(true);", el)
                return SeleniumElement(el, self, self.ToRetryOnError)
            except selenium.common.exceptions.NoSuchElementException as e: 
                if timeout == 0:
                    return None 
                elif timeout == -1:
                    time.sleep(1)
                elif timeout > 0:
                    time.sleep(1)
                    waited += 1
                    if waited > timeout:
                        return None 
            except Exception as e:
                if retryOnError != 0:
                    errcount += 1
                    if errcount > retryOnError:
                        raise e 
                    time.sleep(1)
                else:
                    raise e 

        # import ipdb
        # ipdb.set_trace()
    
    def Exists(self, xpath:str, timeout:int=0) -> bool:
        return self._find(xpath, timeout=timeout) != None
    
    def StatusCode(self) -> int:
        self.driver.stat
    
    def ResizeWindow(self, width:int, height:int):
        """
        :param width: The width of the window in pixels
        :type width: int
        :param height: The height of the window in pixels
        :type height: int
        """
        self.driver.set_window_size(width, height)
    
    def ScrollRight(self, pixel:int):
        """
        ScrollRight(self, pixel:int) scrolls the page to the right by the number of pixels specified in
        the pixel parameter
        
        :param pixel: The number of pixels to scroll by
        :type pixel: int
        """
        self.driver.execute_script("window.scrollBy("+str(pixel)+",0);")
    
    def ScrollLeft(self, pixel:int):
        """
        Scrolls the page left by the number of pixels specified in the parameter.
        
        :param pixel: The number of pixels to scroll by
        :type pixel: int
        """
        self.driver.execute_script("window.scrollBy("+str(pixel*-1)+",0);")

    def ScrollUp(self, pixel:int):
        """
        Scrolls up the page by the number of pixels specified in the parameter.
        
        :param pixel: The number of pixels to scroll up
        :type pixel: int
        """
        self.driver.execute_script("window.scrollBy(0, "+str(pixel*-1)+");")

    def ScrollDown(self, pixel:int):
        """
        Scrolls down the page by the specified number of pixels
        
        :param pixel: The number of pixels to scroll down
        :type pixel: int
        """
        self.driver.execute_script("window.scrollBy(0, "+str(pixel)+");")

    def Url(self) -> str:
        """
        > The `Url()` function returns the current URL of the page
        :return: The current url of the page
        """
        return self.driver.current_url
    
    def Cookie(self) -> list[dict]:
        """
        This function gets the cookies from the driver and returns them as a list of dictionaries
        """
        return self.driver.get_cookies()
    
    def SetCookie(self, cookie:dict|list[dict]):
        """
        If the cookie is a dictionary, add it to the driver. If it's a list of dictionaries, add each
        dictionary to the driver
        
        :param cookie: dict|list[dict]
        :type cookie: dict|list[dict]
        """
        if type(cookie) == dict:
            self.driver.add_cookie(cookie)
        else:
            for i in cookie:
                self.driver.add_cookie(i)
    
    def Refresh(self):
        """
        Refresh() refreshes the current page
        """
        self.driver.refresh()
    
    def GetSession(self) -> str:
        """
        The function GetSession() returns the session ID of the current driver
        :return: The session ID of the driver.
        """
        return self.driver.session_id
    
    @retryOnError
    def Get(self, url:str):
        """
        The function Get() takes a string as an argument and uses the driver object to navigate to the
        url.
        
        :param url: The URL of the page you want to open
        :type url: str
        """

        if self.browserName in ["chrome", "chromewire"]:
            # if self.userAgent != None:
            #     self.driver.execute_cdp_cmd('Network.setUserAgentOverride', {"userAgent": self.userAgent})
            if self.userAgent == None and self.randomUA:
                if not self.browserRemote:
                    self.driver.execute_cdp_cmd('Network.setUserAgentOverride', {"userAgent": random.choice(useragents)['user_agent']})
            

        self.driver.get(url)
    
    def PageSource(self) -> str:
        """
        It returns the page source of the current page
        :return: The page source of the current page.
        """
        return self.driver.page_source

    def Title(self) -> str:
        """
        The function Title() returns the title of the current page
        :return: The title of the page
        """
        return self.driver.title
    
    def Close(self):
        """
        The function closes the browser window and quits the driver
        """
        self.closed = True
        if self.browserName == "chromewire":
            del(self.driver.requests)
        self.driver.close()
        self.driver.quit()
    
    def ClearIdent(self):
        if self.browserName in ["chrome", "chromewire"] :
            try:
                self.driver.delete_all_cookies()
            except:
                pass 
            try:
                self.driver.execute_script("localStorage.clear();")
            except:
                pass 
            try:
                self.driver.execute_script("sessionStorage.clear();")
            except:
                pass 
            try:
                self.driver.execute_script("const dbs = await window.indexedDB.databases();dbs.forEach(db => { window.indexedDB.deleteDatabase(db.name)});")
            except:
                pass
        else:
            raise Exception("未实现")
    
    def ExceptXPath(self, *xpath:str|list, timeout:int=30) -> int | None:
        """
        It waits for some certain elements to appear on the screen.
        
        :param : xpath:str - The xpaths of the element you want to find
        :type : str
        :param timeout: The number of seconds to wait for the element to appear, defaults to 30
        :type timeout: int (optional)
        :return: The index of the xpath that is found.
        """
        if type(xpath[0]) == list:
            xpath = xpath[0]
            
        for _ in range(timeout*2):
            for x in range(len(xpath)):
                if self._find(xpath[x], 0, scrollIntoElement=False) != None:
                    return x
            time.sleep(0.5)

        return None 

    def ExceptKeywords(self, *keywords:str|list, timeout:int=30) -> int | None:
        if type(keywords[0]) == list:
            keywords = keywords[0]
            
        for _ in range(timeout*2):
            for x in range(len(keywords)):
                if keywords[x] in self.PageSource():
                    return x 
                
            time.sleep(0.5)

        return None 
    
    def SwitchTabByID(self, number:int):
        """
        SwitchTabByID(self, number:int) switches to the tab with the given ID, start from 0
        
        :param number: The number of the tab you want to switch to
        :type number: int
        """
        self.driver.switch_to.window(self.driver.window_handles[number])
    
    def SwitchTabByIdent(self, ident:str):
        self.driver.switch_to.window(ident)

    def Tabs(self) -> list[str]:
        return self.driver.window_handles
    
    def NewTab(self) -> str:
        """
        It opens a new tab, and returns the ident of the new tab
        :return: The new tab's ident.
        """
        tabs = self.driver.window_handles
        self.driver.execute_script('''window.open();''')
        for i in self.driver.window_handles:
            if i not in tabs:
                return i
     
    def Screenshot(self, filepath:str) -> str:
        """
        It takes a screenshot of the current page and saves it to the filepath.
        只支持PNG. 
        
        :param filepath: The path to save the screenshot to
        :type filepath: str
        :return: The filepath of the screenshot.
        """
        filepath = Os.Path.Uniquify(filepath)
        self.driver.save_screenshot(filepath)
        return filepath
    
    def MaximizeWindow(self):
        self.driver.maximize_window()
    
    def ExecuteJavaScript(self, code:str) -> typing.Any:
        return self.driver.execute_script(code)
    
    def __enter__(self):
        return self 
    
    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.Close()
        except:
            pass

class Chrome(seleniumBase):
    def __init__(
            self, 
            chromedriverServer:str=None,
            seleniumServer:str=None, 
            PACFileURL:str=None, 
            httpProxy:str=None, 
            sessionID=None, 
            randomUA:bool=True,
            userAgent:str=None,
            # chromeLocation:str=None,
            extensionsPath:str|list=None,
            retryOnError:bool=True,
        ):
        self.ToRetryOnError = retryOnError

        options = webdriver.ChromeOptions()
        # options = chromeoptions()

        if isinstance(extensionsPath, str):
            options.add_extension(extensionsPath)
        elif isinstance(extensionsPath, list):
            for ep in extensionsPath:
                options.add_extension(ep)

        # 这样设置chromedriver的位置会报错:session not created: DevToolsActivePort file doesn't exist
        # 不知为何.
        # 可以这样: driver = webdriver.Chrome(executable_path=chrome_driver_path)
        # 未测试
        # if chromeLocation != None:
        #     options.binary_location = chromeLocation
        # else:
        #     options.binary_location = Cmd.Where("chromedriver")

        # 防止通过navigator.webdriver来检测是否是被selenium操作
        options.add_argument("--disable-blink-features")
        options.add_argument("--disable-blink-features=AutomationControlled")

        if Os.GetUID() == 0:
            options.add_argument("--no-sandbox")
        options.add_argument("--disable-dev-shm-usage")

        options.add_argument("--ignore-certificate-errors")
        options.add_argument('--ignore-ssl-errors')
        # options.add_argument('--no-sandbox')

        if userAgent != None:
            options.add_argument('--user-agent=' + userAgent + '')
        elif randomUA:
            options.add_argument('--user-agent=' + random.choice(useragents)['user_agent'] + '')
        self.randomUA = randomUA
        self.userAgent = userAgent

        options.add_experimental_option("excludeSwitches", ["enable-automation"])

        if PACFileURL:
            options.add_argument("--proxy-pac-url=" + PACFileURL)
        elif httpProxy:
            options.add_argument('--proxy-server=' + httpProxy)

        if chromedriverServer:
            self.driver = webdriver.Remote(
                command_executor=chromedriverServer,
                options=options
            )
        elif seleniumServer:
            if not seleniumServer.endswith("/wd/hub"):
                seleniumServer = seleniumServer + "/wd/hub"
            self.driver = webdriver.Remote(
                command_executor=seleniumServer,
                options=options
            )
        else:
            self.driver = webdriver.Chrome(
                options=options,
            )

        if sessionID:
            self.Close()
            self.driver.session_id = sessionID
        
        self.ResizeWindow(1400, 1000)
        
        self.browserName = "chrome"
        self.browserRemote = seleniumServer != None 
        self.closed = False

# class seleniumFlowRequest():
#     def __init__(self) -> None:
#         self.Time:int = None 
#         self.Headers:dict = None 
#         self.Method:str = None 
#         self.URL:str = None 
#         self.BodyBytes:bytes = b'' 
#         self.Body:str = ''
    
#     def __repr__(self) -> str:
#         body = String(String(self.Body).Repr()).Ommit(80)
#         return f"seleniumFlowRequest(Time={self.Time}, Method={self.Method}, URL={self.URL}, Headers={self.Headers} Body={body})"

#     def __repr__(self) -> str:
#         body = String(String(self.Body).Repr()).Ommit(80)
#         return f"seleniumFlowRequest(Time={self.Time}, Method={self.Method}, URL={self.URL}, Headers={self.Headers} Body={body})"

# class seleniumFlowResponse():
#     def __init__(self) -> None:
#         self.Time:int = None 
#         self.BodyBytes:bytes = b'' 
#         self.Body:str = '' 
#         self.StatusCode:int = None 
#         self.Headers:dict = None 
    
#     def __repr__(self) -> str:
#         body = String(String(self.Body).Repr()).Ommit(80)
#         return f"seleniumFlowResponse(Time={self.Time}, StatusCode={self.StatusCode}, Headers={self.Headers} Body={body})"

#     def __str__(self) -> str:
#         body = String(String(self.Body).Repr()).Ommit(80)
#         return f"seleniumFlowResponse(Time={self.Time}, StatusCode={self.StatusCode}, Headers={self.Headers} Body={body})"

# class SeleniumFlow():
#     def __init__(self,id:str, req:seleniumFlowRequest, resp:seleniumFlowResponse) -> None:
#         self.Request = req
#         self.Response = resp
#         self.ID = id
    
#     def __repr__(self) -> str:
#         return f"SeleniumFlow(\n\tID={self.ID} \n\tRequest={self.Request} \n\tResponse={self.Response}\n)" 

# class ChromeWire(seleniumBase):
#     def __init__(self, 
#             blockSuffix:tuple[str]=seleniumChromeWireSkipFilesSuffix,
#             maxRequests:int=None, 
#             requestStorage:str="memory", # disk
#             urlFilterRegex:list[str]=[], 
#             excludeHosts:list[str]=[],
#             httpProxy:str=None, 
#             sessionID=None, 
#             randomUA:bool=True,
#             userAgent:str=None,
#             chromeLocation:str=None):
        
#         """
#         :param blockSuffix: A tuple of file suffixes to block. list是中括号, tuple是小括号
#         :type blockSuffix: tuple[str]
#         :param maxRequests: The maximum number of requests to store
#         :type maxRequests: int
#         :param requestStorage: The storage type for requests and responses. Can be either memory or disk, defaults to disk
#         :type requestStorage: str (optional)
#         :param urlFilterRegex: A list of regular expressions that will be used to filter requests
#         :type urlFilterRegex: list[str]
#         :param excludeHosts: A list of hosts to exclude from interception
#         :type excludeHosts: list[str]
#         :param httpProxy: The HTTP proxy to use
#         :type httpProxy: str
#         :param sessionID: If you want to use an existing session, you can pass the session ID here
#         :param randomUA: Whether to use a random user agent, defaults to True
#         :type randomUA: bool (optional)
#         """

#         options = webdriver.ChromeOptions()

#         if chromeLocation != None:
#             options.binary_location = chromeLocation
#         else:
#             options.binary_location = Cmd.Where("chromedriver")

#         # 防止通过navigator.webdriver来检测是否是被selenium操作
#         options.add_argument("--disable-blink-features")
#         options.add_argument("--disable-blink-features=AutomationControlled")

#         if Os.GetUID() == 0:
#             options.add_argument("--no-sandbox")
#         options.add_argument("--disable-dev-shm-usage")

#         options.add_argument("--ignore-certificate-errors")

#         if userAgent != None:
#             options.add_argument('--user-agent=' + userAgent + '')
#         elif randomUA:
#             options.add_argument('--user-agent=' + random.choice(useragents)['user_agent'] + '')
#         self.randomUA = randomUA
#         self.userAgent = userAgent

#         options.add_experimental_option("excludeSwitches", ["enable-automation"])

#         seleniumwire_options = {
#             # 'request_storage': 'memory',  # Store requests and responses in memory only
#             # 'request_storage_max_size': maxRequests  # Store no more than 100 requests in memory
#         }

#         if maxRequests != None:
#             seleniumwire_options['request_storage_max_size'] = maxRequests

#         if requestStorage == "memory":
#             seleniumwire_options['request_storage'] = requestStorage
#             if maxRequests == None:
#                 seleniumwire_options['request_storage_max_size'] = 300

#         if excludeHosts != []:
#             seleniumwire_options['exclude_hosts'] = excludeHosts
        
#         if httpProxy != None:
#             seleniumwire_options['proxy'] = {
#                 'http': httpProxy,
#                 # 'https': httpProxy
#             }
        
#         # Lg.Trace(seleniumwire_options)
        
#         self.driver = seleniumwirewebdriver.Chrome(
#             options=options,
#             seleniumwire_options=seleniumwire_options
#         )

#         if self.driver.scopes != []:
#             self.driver.scopes = urlFilterRegex
#             # [
#             #     '.*stackoverflow.*',
#             #     '.*github.*'
#             # ]
        
#         def interceptor(request):
#             # Block PNG, JPEG and GIF images
#             if request.path.lower().endswith(blockSuffix):
#                 request.abort()

#         if blockSuffix != None:
#             self.driver.request_interceptor = interceptor

#         if sessionID:
#             self.Close()
#             self.driver.session_id = sessionID
        
#         self.ResizeWindow(1400, 1000)
        
#         self.browserName = "chromewire"
#         self.browserRemote = False
#         self.closed = False
#         self.blockSuffix = blockSuffix

#         self.tmpCacheFlows = []
    
#     def Flows(self, clean:bool=True) -> typing.Iterable[SeleniumFlow]:
#         """
#         > It iterates through all the requests made by the browser, and returns a `SeleniumFlow` object
#         for each request. 
#         这个方法迭代现有的队列里面的数据, 当到达末尾的时候就结束迭代. 
#         可以多次调用这个方法来获取新的flow. 
#         注意打开某些页面之后, 浏览器会隔一段时间就发起一次请求, 及时没有新Get页面, 也会有新的flow出现. 
#         如果是发起了请求, 但是还在等待回复, 或者正在传送大的回复回来, 则response为None, 在控制台看到的是在pending. 拿到完整的回复才会显示, 一瞬间刷出来.
#         如果是连接出错, 有request, 但response也还是为None. 
#         """
#         while True:
#             Lg.Trace()
#             # for req in self.tmpCacheFlows:
#             if len(self.tmpCacheFlows) == 0:
#                 time.sleep(1)
#                 if len(self.driver.requests) == 0:
#                     break 

#                 self.tmpCacheFlows = copy.deepcopy(self.driver.requests)
#                 Lg.Trace()
#                 if clean:
#                     Lg.Trace()
#                     del(self.driver.requests)
#                 Lg.Trace()
            
#             req = self.tmpCacheFlows.pop(0)

#             if self.blockSuffix != None and len(self.blockSuffix) != 0:
#                 if req.path.lower().endswith(self.blockSuffix):
#                     continue

#             resp = req.response

#             freq = seleniumFlowRequest()
#             freq.URL = req.url
#             freq.Time = req.date.timestamp()
#             freq.Method = req.method
#             freq.Headers = {}
#             for key in req.headers:
#                 freq.Headers[key] = req.headers[key]
#             if req.body != None:
#                 freq.BodyBytes = req.body
#                 freq.Body = req.body.decode(errors="ignore")

#             if resp != None:
#                 fresp = seleniumFlowResponse()
#                 fresp.Headers = {}
#                 for key in resp.headers:
#                     fresp.Headers[key] = resp.headers[key]
#                 fresp.StatusCode = resp.status_code
#                 fresp.Time = resp.date.timestamp()
#                 if resp.body != None:
#                     try:
#                         fresp.BodyBytes = decodeResponseBody(resp.body, resp.headers.get('Content-Encoding', 'identity'))
#                     except Exception as e:
#                         Lg.Warn("解码body报错:", e)
#                         fresp.BodyBytes = b''

#                     fresp.Body = fresp.BodyBytes.decode(errors="ignore")
#                     # if 'content-encoding' not in fresp.Headers:
#                     #     fresp.Body = resp.body.decode(errors="ignore")
#                     #     fresp.BodyBytes = resp.body 
#                     # else:
#                     #     if fresp.Headers['content-encoding'].strip() == 'br':
#                     #         # Lg.Trace("Before:", resp.body[:80])
#                     #         fresp.BodyBytes = brotli.decompress(resp.body)
#                     #         # Lg.Trace("After:", fresp.BodyBytes[:80])
#                     #         fresp.Body = fresp.BodyBytes.decode(errors="ignore")
#                     #     elif fresp.Headers['content-encoding'].strip() == 'gzip':
#                     #         fresp.BodyBytes = gzip.decompress(resp.body)
#                     #         fresp.Body = fresp.BodyBytes.decode(errors="ignore")
#                     #     else:
#                     #         Lg.Warn("未知的 Content-Encoding:", fresp.Headers['content-encoding'])
#                     #         fresp.Body = resp.body.decode(errors="ignore")
#                     #         fresp.BodyBytes = resp.body 
#             else:
#                 fresp = None 

#             Lg.Trace()
#             yield SeleniumFlow(req.id, freq, fresp)

#             if len(self.tmpCacheFlows) == 0:
#                 Lg.Trace()
#                 break 

if __name__ == "__main__":
    # Local 
    # with Chrome() as se:
    # Remote 
    # with Chrome("http://127.0.0.1:4444") as se:

    # With PAC 
    # with Firefox(PACFileURL="http://192.168.1.135:8000/pac") as se:
    # with Chrome("http://127.0.0.1:4444", PACFileURL="http://192.168.1.135:8000/pac") as se:

    # Example of PAC file
    # function FindProxyForURL(url, host)
    # {
    #     if (shExpMatch(host, "*.onion"))
    #     {
    #         return "SOCKS5 192.168.1.135:9150";
    #     }
    #     if (shExpMatch(host, "ipinfo.io"))
    #     {
    #         return "SOCKS5 192.168.1.135:7070";
    #     }
    #     return "DIRECT";
    # }

    # With Proxy
    # with Chrome("http://192.168.1.229:4444", httpProxy="http://192.168.168.54:8899") as se:
        
        # PAC test 
        # se.Get("http://ipinfo.io/ip")
        # print(se.PageSource())

        # se.Get("https://ifconfig.me/ip")
        # print(se.PageSource())
        
        # se.Get("http://juhanurmihxlp77nkq76byazcldy2hlmovfu2epvl5ankdibsot4csyd.onion/")
        # print(se.PageSource())

        # Function test
        # se.Get("https://find-and-update.company-information.service.gov.uk/")
        # inputBar = se.Find("/html/body/div[1]/main/div[3]/div/form/div/div/input")
        # inputBar.Input("ade")
        # button = se.Find('//*[@id="search-submit"]').Click()
        # print(se.PageSource())

    with Chrome(randomUA=False) as se:
        se.Get("http://google.com")
        while True:
            for request in se.Flows():
                ctype = request.Response.Headers['content-encoding'] if 'content-encoding' in request.Response.Headers else "plain"
                Lg.Trace({
                    "url": request.Request.URL,
                    "content-encoding": ctype,
                    "body": request.Response.BodyBytes[:80],
                })

            time.sleep(1)




import os
import urllib.parse as urlparse
import time 
import requests
import tqdm

#print("load " + __file__.split('/')[-1])

def download_file(url, dest):
    with requests.get(url, stream=True) as response:
        response.raise_for_status()
        total_size_in_bytes= int(response.headers.get('content-length', 0))
        chunk_size=8192
        progress_bar = tqdm.tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
        with open(dest, 'wb') as f:
            for chunk in response.iter_content(chunk_size=chunk_size): 
                progress_bar.update(len(chunk))
                f.write(chunk)
        progress_bar.close()
        print(progress_bar.n, total_size_in_bytes)
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            raise Exception("Download error")

def Wget(url:str, dest:str=None, override=True):
    if dest == None:
        fname = os.path.basename(urlparse.urlparse(url).path)
        if len(fname.strip(" \n\t.")) == 0:
            dest = "wget.downloaded." + str(time.time())
        else:
            dest = fname

    if os.path.exists(dest):
        if override:
            download_file(url, dest)
        else:
            raise Exception(f"目标文件已存在: {dest}")
    else:
        download_file(url, dest)

if __name__ == "__main__":
    Wget("http://mirror.nl.leaseweb.net/speedtest/10000mb.bin", "test.10mb")
    import os
    os.system('rm -rf test.10mb')


# -*- coding: utf-8 -*-

from .grammar import SchemaGrammar
from .sqlite_grammar import SQLiteSchemaGrammar
from .postgres_grammar import PostgresSchemaGrammar
from .mysql_grammar import MySQLSchemaGrammar



# -*- coding: utf-8 -*-


class Schema(object):
    def __init__(self, manager):
        """
        :param manager: The database manager
        :type manager: orator.DatabaseManager
        """
        self.db = manager

    def connection(self, connection=None):
        """
        Get a schema builder instance for a connection.

        :param connection: The connection to user
        :type connection: str

        :rtype: orator.schema.SchemaBuilder
        """
        return self.db.connection(connection).get_schema_builder()

    def __getattr__(self, item):
        return getattr(self.db.connection().get_schema_builder(), item)



import sys as __sys
from loguru import logger
import inspect
import os
import threading 
import multiprocessing
import re

pformat = (lambda a:lambda v,t="    ",n="\n",i=0:a(a,v,t,n,i))(lambda f,v,t,n,i:"{%s%s%s}"%(",".join(["%s%s%s: %s"%(n,t*(i+1),repr(k),f(f,v[k],t,n,i+1))for k in v]),n,(t*i)) if type(v)in[dict] else (type(v)in[list]and"[%s%s%s]"or"(%s%s%s)")%(",".join(["%s%s%s"%(n,t*(i+1),f(f,k,t,n,i+1))for k in v]),n,(t*i)) if type(v)in[list,tuple] else repr(v))
logformat = '<green>{time:MM-DD HH:mm:ss}</green> <level>{level:4.4}</level> {message}'

__config = {
    "handlers": [
        {
            "sink": __sys.stdout, 
            # "format": "{time:MM-DD HH:mm:ss} [{icon}] {message}",
            "format": logformat,
            "level": "TRACE",
        },
        # {"sink": "file.log", "serialize": True},
    ],
    # "extra": {"user": "someone"}
}
logger.configure(**__config)

def Trace(*message):
    messages = []
    jstr = " "
    for msg in message:
        if type(msg) == int or type(msg) == float:
            msg = str(msg)
        if type(msg) in [list, dict, set]:
            msg = pformat(msg)
            if msg.count("\n") != 0 and jstr == " ":
                jstr = "\n"
        else:
            msg = str(msg)
        if message.count("\n") != 0 and jstr == " ":
            jstr = "\n"
        messages.append(msg)
    
    p = inspect.stack()[1]

    logger.opt(ansi=True).trace(
        "<cyan>{pname}</cyan>:<cyan>{tname}</cyan>:<cyan>{filename}</cyan>:<cyan>{line}</cyan> <level>{message}</level>", 
        message=jstr.join(messages), 
        function=p.function.replace("<module>", "None"),
        line=p.lineno,
        filename=os.path.basename(p.filename),
        # tid=threading.get_native_id()
        # tid=threading.get_ident(),
        tname=re.sub("\([a-zA-Z0-9]+\)", "", threading.current_thread().name.replace("Thread-", "T").replace(" ", "").replace("MainThread", "MT")),
        pname=multiprocessing.current_process().name.replace("Process-", "P").replace("MainProcess", "MP"),
    )

def Debug(*message):
    messages = []
    jstr = " "
    for msg in message:
        if type(msg) == int or type(msg) == float:
            msg = str(msg)
        if type(msg) in [list, dict, set]:
            msg = pformat(msg)
            if msg.count("\n") != 0 and jstr == " ":
                jstr = "\n"
        else:
            msg = str(msg)
        if message.count("\n") != 0 and jstr == " ":
            jstr = "\n"
        messages.append(msg)
    
    p = inspect.stack()[1]
    
    logger.opt(ansi=True).debug(
        "<cyan>{pname}</cyan>:<cyan>{tname}</cyan>:<cyan>{filename}</cyan>:<cyan>{line}</cyan> <level>{message}</level>", 
        message=jstr.join(messages), 
        function=p.function.replace("<module>", "None"),
        line=p.lineno,
        filename=os.path.basename(p.filename),
        # tid=threading.get_native_id()
        # tid=threading.get_ident(),
        tname=re.sub("\([a-zA-Z0-9]+\)", "", threading.current_thread().name.replace("Thread-", "T").replace(" ", "").replace("MainThread", "MT")),
        pname=multiprocessing.current_process().name.replace("Process-", "P").replace("MainProcess", "MP"),
    )

def Info(*message):
    messages = []
    jstr = " "
    for msg in message:
        if type(msg) == int or type(msg) == float:
            msg = str(msg)
        if type(msg) in [list, dict, set]:
            msg = pformat(msg)
            if msg.count("\n") != 0 and jstr == " ":
                jstr = "\n"
        else:
            msg = str(msg)
        if message.count("\n") != 0 and jstr == " ":
            jstr = "\n"
        messages.append(msg)
    
    p = inspect.stack()[1]
    
    logger.opt(ansi=True).info(
        "<cyan>{pname}</cyan>:<cyan>{tname}</cyan>:<cyan>{filename}</cyan>:<cyan>{line}</cyan> <level>{message}</level>", 
        message=jstr.join(messages), 
        function=p.function.replace("<module>", "None"),
        line=p.lineno,
        filename=os.path.basename(p.filename),
        # tid=threading.get_native_id()
        # tid=threading.get_ident(),
        tname=re.sub("\([a-zA-Z0-9]+\)", "", threading.current_thread().name.replace("Thread-", "T").replace(" ", "").replace("MainThread", "MT")),
        pname=multiprocessing.current_process().name.replace("Process-", "P").replace("MainProcess", "MP"),
    )

def Warn(*message):
    messages = []
    jstr = " "
    for msg in message:
        if type(msg) == int or type(msg) == float:
            msg = str(msg)
        if type(msg) in [list, dict, set]:
            msg = pformat(msg)
            if msg.count("\n") != 0 and jstr == " ":
                jstr = "\n"
        else:
            msg = str(msg)
        if message.count("\n") != 0 and jstr == " ":
            jstr = "\n"
        messages.append(msg)
    
    p = inspect.stack()[1]
    
    logger.opt(ansi=True).warning(
        "<cyan>{pname}</cyan>:<cyan>{tname}</cyan>:<cyan>{filename}</cyan>:<cyan>{line}</cyan> <level>{message}</level>", 
        message=jstr.join(messages), 
        function=p.function.replace("<module>", "None"),
        line=p.lineno,
        filename=os.path.basename(p.filename),
        # tid=threading.get_native_id()
        # tid=threading.get_ident(),
        tname=re.sub("\([a-zA-Z0-9]+\)", "", threading.current_thread().name.replace("Thread-", "T").replace(" ", "").replace("MainThread", "MT")),
        pname=multiprocessing.current_process().name.replace("Process-", "P").replace("MainProcess", "MP"),
    )

def Error(*message, exc:bool=True):
    """
    It logs the error message with the file name, line number, thread name, and process name.
    
    :param exc: If True, the exception will be logged, defaults to False
    :type exc: bool (optional)
    """
    messages = []
    jstr = " "
    for msg in message:
        if type(msg) == int or type(msg) == float:
            msg = str(msg)
        if type(msg) in [list, dict, set]:
            msg = pformat(msg)
            if msg.count("\n") != 0 and jstr == " ":
                jstr = "\n"
        else:
            msg = str(msg)
        if message.count("\n") != 0 and jstr == " ":
            jstr = "\n"
        messages.append(msg)
    
    p = inspect.stack()[1]
    
    if exc:
        logger.opt(ansi=True).exception(
            "<cyan>{pname}</cyan>:<cyan>{tname}</cyan>:<cyan>{filename}</cyan>:<cyan>{line}</cyan> <level>{message}</level>", 
            message=jstr.join(messages), 
            function=p.function.replace("<module>", "None"),
            line=p.lineno,
            filename=os.path.basename(p.filename),
            # tid=threading.get_native_id()
            # tid=threading.get_ident(),
            tname=re.sub("\([a-zA-Z0-9]+\)", "", threading.current_thread().name.replace("Thread-", "T").replace(" ", "").replace("MainThread", "MT")),
            pname=multiprocessing.current_process().name.replace("Process-", "P").replace("MainProcess", "MP"),
        )
    else:
        logger.opt(ansi=True).error(
            "<cyan>{pname}</cyan>:<cyan>{tname}</cyan>:<cyan>{filename}</cyan>:<cyan>{line}</cyan> <level>{message}</level>", 
            message=jstr.join(messages), 
            function=p.function.replace("<module>", "None"),
            line=p.lineno,
            filename=os.path.basename(p.filename),
            # tid=threading.get_native_id()
            # tid=threading.get_ident(),
            tname=re.sub("\([a-zA-Z0-9]+\)", "", threading.current_thread().name.replace("Thread-", "T").replace(" ", "").replace("MainThread", "MT")),
            pname=multiprocessing.current_process().name.replace("Process-", "P").replace("MainProcess", "MP"),
        )

def SetLevel(level: str):
    """
    It sets the logging level of the logger to the level passed in
    
    :param level: The level of messages to log. canbe: trace,debug,info,warn,error
    :type level: str
    """
    for idx in range(len(__config['handlers'])):
        __config['handlers'][idx]['level'] = level.upper()

    logger.configure(**__config)

def SetStdout(enable:bool):
    if enable == False:
        handlers = []
        for idx in range(len(__config['handlers'])):
            if __config['handlers'][idx]['sink'] != __sys.stdout:
                handlers.append(__config['handlers'][idx])
        
        __config['handlers'] = handlers
    else:
        if __sys.stdout not in [i["sink"] for i in __config['handlers']]:
            handler = {
                "sink": __sys.stdout, 
                "format": logformat,
                "level": "TRACE",
            }
            __config['handlers'].append(handler)

    logger.configure(**__config)

def SetFile(path:str, size:int=100, during:int=7, color:bool=True, json:bool=False):
    """
    It sets the file handler for the logger.
    
    :param path: The path to the log file
    :type path: str
    :param size: The size of the file before it rotates, in MB
    :type size: int
    :param during: how long to keep the log file, in Hour
    :type during: int
    :param color: If True, the output will be colorized, defaults to True
    :type color: bool (optional)
    :param json: If True, the log records will be serialized to JSON, defaults to False
    :type json: bool (optional)
    """
    if '/' in path.strip("/") and not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path.strip('/')))

    if path not in [i["sink"] for i in __config['handlers']]:
        handler = {
            "sink": path,
            "rotation": str(size)+" MB", 
            "retention": str(during)+" hours", 
            "format": logformat,
            "level": __config['handlers'][0]['level'] if len(__config['handlers']) > 0 else "TRACE",
            "colorize": color,
            "serialize": json,
        }
        __config['handlers'].append(handler)
        logger.configure(**__config)

# import time 

# def ff():    
#     def f():
#         while True:
#             time.sleep(1)
#             Trace(time.time())

#     t = threading.Thread(target=f)
#     t.daemon = True 
#     t.start()

#     time.sleep(99999)

if __name__ == "__main__":
    # SetLevel("info")
    # SetFile("test.log", 1, 1, json=True)
    Trace(True)
    Trace("trace")
    Debug("debug")
    Info("info")
    Warn("warn")
    Warn(False)
    Error("error")
    Debug("text debug message", [ ['spam', 'eggs', 'lumberjack', 'knights', 'ni'], 'spam', 'eggs', 'lumberjack', 'knights', 'ni'])
    Trace("text debug message", [ ['spam', 'eggs', 'lumberjack', 'knights', 'ni'], 'spam', 'eggs', 'lumberjack', 'knights', 'ni'])
    Debug("first", "second", "third")
    Trace("初始化实例", 1)
    try:
        int("2.3")
    except:
        Error("转换错误:", True)

    
    # p = multiprocessing.Process(target=ff)
    # #p.daemon = True 
    # p.start()

    # time.sleep(99999)



import nslookup
from dns import resolver
import dns.reversename
from .. import Random
import dns

#print("load " + '/'.join(__file__.split('/')[-2:]))

class Nslookup():
    def __init__(self, server:list[str]=["8.8.8.8", "1.1.1.1", "8.8.4.4"], tcp:bool=False) -> None:
        if type(server) == str:
            server = [server]

        self.server = server
        self.tcp = tcp
    
    def A(self, domain:str) -> list[str]:
        nl = nslookup.Nslookup(dns_servers=Random.Choice(self.server, 5), tcp=self.tcp)
        return nl.dns_lookup(domain).answer
    
    def AAAA(self, domain:str) -> list[str]:
        nslookup = nslookup.Nslookup(dns_servers=Random.Choice(self.server, 5), tcp=self.tcp)
        return nslookup.dns_lookup6(domain).answer
    
    def Reverse(self, ip:str) -> str:
        rl = resolver.Resolver()
        rl.nameservers = Random.Choice(self.server, 5)
        return str(resolver.resolve(dns.reversename.from_address(ip), "PTR")[0])
    
    def MX(self, domain:str) -> list[str]:
        rl = resolver.Resolver()
        rl.nameservers = Random.Choice(self.server, 5)

        try:
            answer = rl.resolve(domain, "MX")
        except dns.resolver.NoAnswer:
            return []

        res = {}
        for x in answer:
            x = [i.strip() for i in x.to_text().split()]
            res[int(x[0])] = x[1]
        
        p = list(res)
        p.sort()

        ress = []
        for pp in p:
            ress.append(res[pp])
        
        return ress


# -*- coding: utf-8 -*-


class QueryExpression(object):
    def __init__(self, value):
        self._value = value

    def get_value(self):
        return self._value

    def __str__(self):
        return str(self.get_value())



from __future__ import annotations

import schedule
import time 

from ..Thread import Thread

#print("load " + '/'.join(__file__.split('/')[-2:]))

class Crontab():
    def __init__(self):
        Thread(self.run)

    def run(self):
        while True:
            schedule.run_pending()
            time.sleep(1) 

    def Every(self, interval: int = 1) -> Crontab:
        self.obj = schedule.every(interval) 
        self.everyInterval = interval
        return self
    
    def Second(self) -> Crontab:
        if self.everyInterval == 1:
            self.obj = self.obj.second
        elif self.everyInterval > 1:
            self.obj = self.obj.seconds

        return self 
    
    def Minute(self) -> Crontab:
        if self.everyInterval == 1:
            self.obj = self.obj.minute
        elif self.everyInterval > 1:
            self.obj = self.obj.minutes

        return self 
    
    def Hour(self) -> Crontab:
        if self.everyInterval == 1:
            self.obj = self.obj.hour
        elif self.everyInterval > 1:
            self.obj = self.obj.hours

        return self 
    
    def Day(self) -> Crontab:
        if self.everyInterval == 1:
            self.obj = self.obj.day
        elif self.everyInterval > 1:
            self.obj = self.obj.days

        return self 

    def Week(self) -> Crontab:
        if self.everyInterval == 1:
            self.obj = self.obj.week
        elif self.everyInterval > 1:
            self.obj = self.obj.weeks

        return self 
    
    def At(self, time: str) -> Crontab:
        self.obj = self.obj.at(time)
        return self

    def Do(self, job_func, *args, **kwargs):
        self.obj.do(job_func, *args, **kwargs)
    
    def Monday(self):
        self.obj = self.obj.monday 
        return self 
    
    def Tuesday(self):
        self.obj = self.obj.tuesday 
        return self 
    
    def Wednesday(self):
        self.obj = self.obj.wednesday 
        return self  

    def Thursday(self):
        self.obj = self.obj.thursday 
        return self 
    
    def Friday(self):
        self.obj = self.obj.friday 
        return self 
    
    def Saturday(self):
        self.obj = self.obj.saturday 
        return self 
    
    def Sunday(self):
        self.obj = self.obj.sunday 
        return self 

if __name__ == "__main__":
    def job():
        print("I'm working...")
    
    c = Crontab()

    c.Every(3).Second().Do(job)
    c.Every(3).Minute().Do(job)
    c.Every(3).Hour().Do(job)
    c.Every(3).Day().Do(job)
    c.Every(3).Week().Do(job)

    # Run job every minute at the 23rd second
    c.Every().Minute().At(":23").Do(job)

    # Run job every hour at the 42rd minute
    c.Every().Hour().At(":42").Do(job)

    # Run jobs every 5th hour, 20 minutes and 30 seconds in.
    # If current time is 02:00, first execution is at 06:20:30
    c.Every(5).Hour().At("20:30").Do(job)

    # Run job every day at specific HH:MM and next HH:MM:SS
    c.Every().Day().At("10:30").Do(job)
    c.Every().Day().At("10:30:42").Do(job)

    # Run job on a specific day of the week
    c.Every().Monday().Do(job)
    c.Every().Wednesday().At("13:15").Do(job)
    c.Every().Minute().At(":17").Do(job)

    while True:
        time.sleep(1)


import os
import sys 
import re

#print("load os")

Stdin = sys.stdin
Stdout = sys.stdout

def Touch(path:str):
    from pathlib import Path
    Path(path).touch()

def Chdir(path:str):
    os.chdir(path)

def Exit(num:int=0):
    sys.exit(num)

def System(cmd:str) -> int:
    import subprocess
    return subprocess.call(cmd, stderr=sys.stderr, stdout=sys.stdout, shell=True)

def Mkdir(*path:str):
    for p in path:
        os.makedirs(p, exist_ok=True)

def ListDir(path:str=".") -> list[str]:
    return os.listdir(path)

def ListFiles(path:str) -> list[str]:
    import glob
    return glob.glob(path)

Args = sys.argv 

def Getenv(varname:str, defaultValue:str=None) -> str | None:
    v = os.environ.get(varname)
    if not v:
        return defaultValue
    else:
        return v

def Getcwd() -> str:
    return os.getcwd()

def Unlink(path:str):
    import shutil
    if os.path.exists(path):
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.unlink(path)

def Move(src:str, dst:str, force:bool=True):
    import shutil
    if os.path.exists(dst):
        if not os.path.isdir(dst):
            if not force:
                raise Exception("目标已存在")
            else:
                os.unlink(dst)
        else:
            dst = os.path.join(dst, os.path.basename(src))

    ddir = os.path.dirname(dst)
    if ddir != "":
        if not os.path.exists(ddir):
            Mkdir(ddir)
    
    shutil.move(src, dst)

def Copy(src:str, dst:str, force:bool=True):
    import shutil
    if os.path.exists(dst):
        if not os.path.isdir(dst):
            if not force:
                raise Exception("目标已存在")
            else:
                os.unlink(dst)
        else:
            dst = os.path.join(dst, os.path.basename(src))
    
    ddir = os.path.dirname(dst)
    if ddir != "":
        if not os.path.exists(ddir):
            Mkdir(ddir)
    
    shutil.copy2(src, dst)

def GetLoginUserName() -> str:
    return os.getlogin()

def Walk(path:str, type:str=None) -> str:
    """
    Walk through a directory and yield the names.
    
    :param path: The path to the directory you want to walk
    :type path: str
    :param type: The type of file you want to search for. "d" for directory and "f" for file, None(default) for all
    :type type: str
    """
    for root, dirs, files in os.walk(path, topdown=False):
        if type == None:
            for name in files:
                yield os.path.join(root, name)
            for name in dirs:
                yield os.path.join(root, name)
        elif type == "f":
            for name in files:
                yield os.path.join(root, name)
        elif type == "d":
            for name in dirs:
                yield os.path.join(root, name)

def GetUID() -> int:
    return os.getuid()

def GetCurrentThreadID() -> str:
    import threading
    import multiprocessing
    return multiprocessing.current_process().name.replace("Process-", "P").replace("MainProcess", "MP") + re.sub("\([a-zA-Z0-9]+\)", "", threading.current_thread().name.replace("Thread-", "T").replace(" ", "").replace("MainThread", "MT")) 

def GetIPByInterface(interface:str) -> str:
    import psutil
    import socket

    addrs = psutil.net_if_addrs()
    if interface in addrs:
        for addr in addrs[interface]:
            if addr.family == socket.AF_INET:  # 使用 socket.AF_INET
                return addr.address
            
    raise Exception("No IP address found for this interface")

if __name__ == "__main__":
    # Move("a", "b") # 移动当前目录的a到b
    # Move("b", "c/d/e") # 移动b到c/d/e, 会先递归创建目录c/d
    # Move("c/d/e", "d") # 移动c/d/e文件到d目录, 没有指定文件名就自动使用原来的文件名
    for i in Walk(".", type="d"):
        print(i)